// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of demoHeterogeneousParallelism.
// 
// demoHeterogeneousParallelism is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// demoHeterogeneousParallelism is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
// even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
// 
// You should have received a copy of the GNU General Public License along with demoHeterogeneousParallelism. If not, 
// see <http://www.gnu.org/licenses/>.


// place-holder for your code: difference of gaussians sample application
// ======================================================================

#include "DoG.h"
#include "hawaii/GPU/pool.h"
#include "hawaii/common/error.h"

// gaussian blur: Results are cropped by window size minus one to avoid border extrapolation. Window covers five sigma.
void gaussCPU( const cv::Mat  orig,
                     cv::Mat& blur,
               const cv::Size win  ) {
	
	// check input
	HAWAII_ERROR_CONDITIONAL( orig.rows < win.height
	                       || orig.cols < win.width,
	                          "input too small for window" ) ;
	HAWAII_ERROR_CONDITIONAL( !( win.height & 1 )
	                       || !( win.width  & 1 ),
	                          "Window dimensions must be odd." ) ;
	// wrap OpenCV
	// developer note: The CPU version realizes that the input image goes beyond the given ROI and therefore does not 
	//                 perform any border extrapolation.
	cv::GaussianBlur( orig( cv::Range( win.height / 2, orig.rows - win.height / 2 ),
	                        cv::Range( win.width  / 2, orig.cols - win.width  / 2 ) ),
	                  blur,
	                  win,
	                  win.width / 5.0, win.height / 5.0,
	                  cv::BORDER_DEFAULT ) ;
}
void gaussGPU( const cv::gpu::GpuMat  orig,
                     cv::gpu::GpuMat& blur,
               const cv::Size         win,
                     cv::gpu::Stream& stream ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( orig.rows < win.height
	                       || orig.cols < win.width,
	                          "input too small for window" ) ;
	HAWAII_ERROR_CONDITIONAL( !( win.height & 1 )
	                       || !( win.width  & 1 ),
	                          "Window dimensions must be odd." ) ;
	
	// select ROI, allocate buffer and result from pool to avoid device synchronization
	cv::gpu::GpuMat blurFull, buffer ;
	hawaii::GPU::memoryPool( blurFull, orig.size(), orig.type()                            ) ;
	hawaii::GPU::memoryPool( buffer,   orig.size(), CV_MAKETYPE( CV_32F, orig.channels() ) ) ;
	
	// wrap OpenCV
	// developer note #1: The GPU version does not realize if the input image goes beyond the given ROI, therefore one 
	//                    must first calculate all result pixels including borders and then crop them.
	// developer note #2: The OpenCV 3.0 interface does not even support an externally allocated buffer and therefore is 
	//                    not asynchronous. But even the 2.4 interface with a buffer is not asynchronous.
	#if CV_MAJOR_VERSION == 3
		auto cvGaussianFilterGPU = cv::gpu::createGaussianFilter( CV_8UC1, CV_8UC1, win, 0 ) ;
		cvGaussianFilterGPU->apply( orig,
		                            blurFull,
		                            stream   ) ;
	#elif CV_MAJOR_VERSION == 2
		cv::gpu::GaussianBlur( orig,
		                       blurFull,
		                       win,
		                       buffer,
		                       win.width / 5.0, win.height / 5.0,
		                       cv::BORDER_DEFAULT, cv::BORDER_DEFAULT,
		                       stream                                 ) ;
	#else
		#error "unsupported version of OpenCV"
	#endif
	blur = blurFull( cv::Range( win.height / 2, blurFull.rows - win.height / 2 ),
	                 cv::Range( win.width  / 2, blurFull.cols - win.width  / 2 ) ) ;
}

// difference between images: auto-select center ROI if "orig" is bigger than "blur" due to cropping, map result range 
//                            from [-255,255] to [0,255]
void diffCPU( const cv::Mat  orig,
              const cv::Mat  blur,
                    cv::Mat& diff ) {
	
	// find size difference
	const int cropVerti2 = ( orig.rows - blur.rows ) / 2,
	          cropHoriz2 = ( orig.cols - blur.cols ) / 2 ;
	
	// select ROI
	const cv::Mat origROI = orig( cv::Range( cropVerti2, cropVerti2 + blur.rows ),
	                              cv::Range( cropHoriz2, cropHoriz2 + blur.cols ) ) ;
	// use OpenCV
	cv::addWeighted( origROI,  0.5,
	                 blur,    -0.5,
	                 127.5,
	                 diff,
	                 -1 ) ;
}
void diffGPU( const cv::gpu::GpuMat  orig,
              const cv::gpu::GpuMat  blur,
                    cv::gpu::GpuMat& diff,
                    cv::gpu::Stream& stream ) {
	
	// find size difference
	const int cropVerti2 = ( orig.rows - blur.rows ) / 2,
	          cropHoriz2 = ( orig.cols - blur.cols ) / 2 ;
	
	// select ROI, allocate result from pool to avoid device synchronization
	const cv::gpu::GpuMat origROI = orig( cv::Range( cropVerti2, cropVerti2 + blur.rows ),
	                                      cv::Range( cropHoriz2, cropHoriz2 + blur.cols ) ) ;
	hawaii::GPU::memoryPool( diff, origROI.size(), origROI.type() ) ;
	
	// use OpenCV
	cv::gpu::addWeighted( origROI,  0.5,
	                      blur,    -0.5,
	                      127.5,
	                      diff,
	                      -1,
	                      stream ) ;
}
