// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of demoHeterogeneousParallelism.
// 
// demoHeterogeneousParallelism is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// demoHeterogeneousParallelism is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
// even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
// 
// You should have received a copy of the GNU General Public License along with demoHeterogeneousParallelism. If not, 
// see <http://www.gnu.org/licenses/>.


// "libHawaii" processing steps to wrap existing code: difference of gaussians sample application
// ==============================================================================================

#pragma once

#include "hawaii/flow/itemNamedImages.h"
#include "hawaii/flow/step.h"
#include "hawaii/flow/stepGPU.h"
#include "DoG.h"
#include "ompWrapper.h"

// gaussian blur on CPU
class GaussCPU : public hawaii::flow::StepTemplate< hawaii::flow::ItemNamedImages,
                                                    hawaii::flow::ItemNamedImages > {
	// options as members, set name and cores used
	public:
	cv::Size win ;
	GaussCPU() :
		win( 11, 11 ) {
		this->name = "CPU blur" ;
		this->setCPUCores( omp_get_max_threads() ) ;
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// pass work item through
		dstPtr = srcPtr ;
		
		// wrap CPU function from "DoG.h"
		gaussCPU( srcPtr->images[ "orig" ],
		          dstPtr->images[ "blur" ],
		          this->win                ) ;
	}
} ;

// gaussian blur on GPU
class GaussGPU : public hawaii::flow::StepTemplateGPU< hawaii::flow::ItemNamedImages,
                                                       hawaii::flow::ItemNamedImages > {
	// options as members, set name
	public:
	cv::Size win ;
	GaussGPU() :
		win( 11, 11 ) {
		this->name = "GPU blur" ;
	}
	
	// inherited purely-virtual methods
	public:
	void uploadImpl( const SrcPtr&          srcPtr,
	                       SrcGPUPtr&       srcGPUPtr,
	                       cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		srcGPUPtr = srcPtr ;
		
		// upload input to GPU
		srcGPUPtr->images[ "orig" ].readGPU( stream ) ;
	}
	void computeImpl( const SrcGPUPtr&       srcGPUPtr,
	                        DstGPUPtr&       dstGPUPtr,
	                        cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstGPUPtr = srcGPUPtr ;
		
		// wrap GPU function from "DoG.h"
		gaussGPU( srcGPUPtr->images[ "orig" ],
		          dstGPUPtr->images[ "blur" ],
		          this->win,
		          stream                      ) ;
	}
	void downloadImpl( const DstGPUPtr&       dstGPUPtr,
	                         DstPtr&          dstPtr,
	                         cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstPtr = dstGPUPtr ;
		
		// download result from GPU
		dstPtr->images[ "blur" ].readCPU( stream ) ;
	}
} ;

// difference between images on CPU
class DiffCPU : public hawaii::flow::StepTemplate< hawaii::flow::ItemNamedImages,
                                                   hawaii::flow::ItemNamedImages > {
	// set name and cores used
	public:
	DiffCPU() {
		this->name = "CPU diff" ;
		this->setCPUCores( omp_get_max_threads() ) ;
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// pass work item through
		dstPtr = srcPtr ;
		
		// wrap CPU function from "DoG.h"
		diffCPU( srcPtr->images[ "orig" ],
		         srcPtr->images[ "blur" ],
		         dstPtr->images[ "diff" ] ) ;
	}
} ;

// difference between images on GPU
class DiffGPU : public hawaii::flow::StepTemplateGPU< hawaii::flow::ItemNamedImages,
                                                      hawaii::flow::ItemNamedImages > {
	// set name
	public:
	DiffGPU() {
		this->name = "GPU diff" ;
	}
	
	// inherited purely-virtual methods
	public:
	void uploadImpl( const SrcPtr&          srcPtr,
	                       SrcGPUPtr&       srcGPUPtr,
	                       cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		srcGPUPtr = srcPtr ;
		
		// upload inputs to GPU
		srcGPUPtr->images[ "orig" ].readGPU( stream ) ;
		srcGPUPtr->images[ "blur" ].readGPU( stream ) ;
	}
	void computeImpl( const SrcGPUPtr&       srcGPUPtr,
	                        DstGPUPtr&       dstGPUPtr,
	                        cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstGPUPtr = srcGPUPtr ;
		
		// wrap GPU function from "DoG.h"
		diffGPU( srcGPUPtr->images[ "orig" ],
		         srcGPUPtr->images[ "blur" ],
		         dstGPUPtr->images[ "diff" ],
		         stream ) ;
	}
	void downloadImpl( const DstGPUPtr&       dstGPUPtr,
	                         DstPtr&          dstPtr,
	                         cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstPtr = dstGPUPtr ;
		
		// download result from GPU
		dstPtr->images[ "diff" ].readCPU( stream ) ;
	}
} ;

// both processing steps on GPU
// user note: Because "hawaii::flow::Series" would disable GPU latency hiding via overlapping memory copies and 
//            computations, this specific combination is implemented manually.
class GaussDiffGPU : public hawaii::flow::StepTemplateGPU< hawaii::flow::ItemNamedImages,
                                                           hawaii::flow::ItemNamedImages > {
	// options as members, set name
	public:
	cv::Size win ;
	GaussDiffGPU() :
		win( 11, 11 ) {
		this->name = "GPU blur and diff" ;
	}
	
	// inherited purely-virtual methods
	public:
	void uploadImpl( const SrcPtr&          srcPtr,
	                       SrcGPUPtr&       srcGPUPtr,
	                       cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		srcGPUPtr = srcPtr ;
		
		// upload input to GPU
		srcGPUPtr->images[ "orig" ].readGPU( stream ) ;
	}
	void computeImpl( const SrcGPUPtr&       srcGPUPtr,
	                        DstGPUPtr&       dstGPUPtr,
	                        cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstGPUPtr = srcGPUPtr ;
		
		// wrap GPU functions from "DoG.h"
		gaussGPU( srcGPUPtr->images[ "orig" ],
		          dstGPUPtr->images[ "blur" ],
		          this->win,
		          stream                      ) ;
		diffGPU( srcGPUPtr->images[ "orig" ],
		         dstGPUPtr->images[ "blur" ],
		         dstGPUPtr->images[ "diff" ],
		         stream                      ) ;
	}
	void downloadImpl( const DstGPUPtr&       dstGPUPtr,
	                         DstPtr&          dstPtr,
	                         cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstPtr = dstGPUPtr ;
		
		// download results from GPU
		dstPtr->images[ "blur" ].readCPU( stream ) ;
		dstPtr->images[ "diff" ].readCPU( stream ) ;
	}
} ;
