// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of demoHeterogeneousParallelism.
// 
// demoHeterogeneousParallelism is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// demoHeterogeneousParallelism is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
// even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
// 
// You should have received a copy of the GNU General Public License along with demoHeterogeneousParallelism. If not, 
// see <http://www.gnu.org/licenses/>.


// place-holder for your code: GEMM sample application
// ===================================================

#include "GEMM.h"
#include "hawaii/GPU/CUDA.h"
#include "hawaii/GPU/pool.h"
#include "hawaii/common/partitionize.h"
#include "hawaii/common/error.h"
#include <eigen3/Eigen/Dense>
#include <opencv2/core/eigen.hpp>
#ifdef _OPENMP
	#include <omp.h>
#else
	static int omp_get_max_threads() { return 1 ; }
#endif

// helpers only used here
namespace {
	
	// wrap OpenCV matrices as Eigen matrices
	// developer note: "cv::cv2eigen()" produced empty ones.
	typedef Eigen::Map< Eigen::Matrix< float, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor >, 0, Eigen::OuterStride<> > EigenMap ;
	EigenMap cvMat2EigenMap( cv::Mat cvMat ) {
		assert( cvMat.type() == CV_32FC1 ) ;
		return EigenMap( (float*)cvMat.data, cvMat.rows, cvMat.cols, Eigen::OuterStride<>( cvMat.step / sizeof( float ) ) ) ;
	}
	
	// adapted from "cv::gpu::gemm()", which claims to be asynchronous, but newly initializes CUBLAS on every call and
	// hereby causes device synchronization by allocating GPU memory
	void GEMMCUBLASWrapper( const cv::gpu::GpuMat  A,
	                        const cv::gpu::GpuMat  B,
	                        const float            alpha,
	                        const cv::gpu::GpuMat  C,
	                        const float            beta,
	                              cv::gpu::GpuMat& result,
	                              cublasHandle_t   handle,
	                              cv::gpu::Stream& stream ) {
		
		// check compile-time GPU support
		#ifdef HAWAII_NO_CUDA
			HAWAII_ERROR_NO_CUDA ;
			(void)A ;
			(void)B ;
			(void)C ;
			(void)alpha ;
			(void)beta  ;
			(void)result ;
			(void)handle ;
			(void)stream ;
		#else
			
			// fall-back only: initialize cuBLAS just for this call
			const bool temporaryHandle = ( handle == nullptr ) ;
			if( temporaryHandle ) { cublasCreate_v2( &handle ) ; }
			
			// allocate result
			hawaii::GPU::memoryPool( result, A.rows, B.cols, A.type() ) ;
			
			// workaround for CUBLAS's "C == result" interface
			if( beta != 0 ) {
				if( C.empty() ) { hawaii::GPU::assignAsync( result, cv::Scalar::all( 0 ), stream ) ; }
				else { hawaii::GPU::copyAsync( C, result, stream ) ; }
			}
			
			// options for subsequent cuBLAS calls
			HAWAII_ERROR_CHECK_CUBLAS( cublasSetStream_v2( handle, cv::gpu::StreamAccessor::getStream( stream ) ) ) ;
			HAWAII_ERROR_CHECK_CUBLAS( cublasSetPointerMode_v2( handle, CUBLAS_POINTER_MODE_HOST ) ) ;
			
			// do the computation: exchanged arguments because OpenCV types are row-major, but cuBLAS is column-major
			HAWAII_ERROR_CHECK_CUBLAS( cublasSgemm_v2( handle,
												CUBLAS_OP_N, CUBLAS_OP_N,
												B.cols, A.rows, B.rows,
												&alpha,
												B.ptr< float >(), static_cast< int >( B.step / sizeof( float ) ),
												A.ptr< float >(), static_cast< int >( A.step / sizeof( float ) ),
												&beta,
												result.ptr< float >(), static_cast< int >( result.step / sizeof( float ) ) ) ) ;
			
			// fall-back only: de-initialize cuBLAS 
			if( temporaryHandle ) { cublasDestroy_v2( handle ) ; }
			
		#endif // #ifdef-#else "compile-time GPU support?"
		
	} // function "GEMMCUBLASWrapper()"
	
} // anonymous namespace

// scaled matrix multiplication: result = A * B * alpha
void multiplyCPU( const cv::Mat  A,
                  const cv::Mat  B,
                        cv::Mat& result ) {
	
	// check input
	HAWAII_ERROR_CONDITIONAL( A.cols   != B.rows,
	                          "size mismatch" ) ;
	HAWAII_ERROR_CONDITIONAL( A.type() != CV_32FC1
	                       || B.type() != CV_32FC1,
	                          "type must be single-channel float" ) ;
	// allocate result
	result.create( A.rows, B.cols, A.type() ) ;
	
	// wrap Eigen, partition to use multiple cores
	const auto BFull = cvMat2EigenMap( B ) ;
	const std::vector< cv::Range > ranges = hawaii::partitionize( result.rows, omp_get_max_threads() ) ;
	#pragma omp parallel for
	for( size_t part = 0 ; part < ranges.size() ; ++part ) {
		
		// extract and wrap partitions
		const auto      APart = cvMat2EigenMap(      A.rowRange( ranges[ part ] ) ) ;
		      auto resultPart = cvMat2EigenMap( result.rowRange( ranges[ part ] ) ) ;
		
		// calculate product
		resultPart = APart * BFull ;
	}
}
void multiplyGPU( const cv::gpu::GpuMat  A,
                  const cv::gpu::GpuMat  B,
                        cv::gpu::GpuMat& result,
                        cublasHandle_t   handle,
                        cv::gpu::Stream& stream ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( A.cols   != B.rows,
	                          "size mismatch" ) ;
	HAWAII_ERROR_CONDITIONAL( A.type() != CV_32FC1
	                       || B.type() != CV_32FC1,
	                          "type must be single-channel float" ) ;
	// allocate result
	hawaii::GPU::memoryPool( result, A.rows, B.cols, A.type() ) ;
	
	// wrap OpenCV
	GEMMCUBLASWrapper( A, B, 1.0f, cv::gpu::GpuMat(), 0.0f, result, handle, stream ) ;
}

// scaled matrix addition: result = AB * alpha + C * beta
void addScaledCPU( const cv::Mat  AB,
                   const cv::Mat  C,
                         cv::Mat& result,
                   const float    alpha,
                   const float    beta   ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( AB.size() != C.size(),
	                          "size mismatch" ) ;
	HAWAII_ERROR_CONDITIONAL( AB.type() != CV_32FC1
	                       ||  C.type() != CV_32FC1,
	                          "type must be single-channel float" ) ;
	// allocate result
	result.create( AB.rows, AB.cols, AB.type() ) ;
	
	// wrap Eigen, partition to use multiple cores
	const std::vector< cv::Range > ranges = hawaii::partitionize( C.rows, omp_get_max_threads() ) ;
	#pragma omp parallel for
	for( size_t part = 0 ; part < ranges.size() ; ++part ) {
		
		// extract and wrap partitions
		const auto     ABPart = cvMat2EigenMap(     AB.rowRange( ranges[ part ] ) ) ;
		const auto      CPart = cvMat2EigenMap(      C.rowRange( ranges[ part ] ) ) ;
		      auto resultPart = cvMat2EigenMap( result.rowRange( ranges[ part ] ) ) ;
		
		// calculate weighted sum
		resultPart = ABPart * alpha + CPart * beta ;
	}
}
void addScaledGPU( const cv::gpu::GpuMat  AB,
                   const cv::gpu::GpuMat  C,
                         cv::gpu::GpuMat& result,
                   const float            alpha,
                   const float            beta,
                         cv::gpu::Stream& stream ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( AB.size() != C.size(),
	                          "size mismatch" ) ;
	HAWAII_ERROR_CONDITIONAL( AB.type() != CV_32FC1
	                       ||  C.type() != CV_32FC1,
	                          "type must be single-channel float" ) ;
	// allocate result
	hawaii::GPU::memoryPool( result, AB.rows, AB.cols, AB.type() ) ;
	
	// wrap OpenCV
	cv::gpu::addWeighted( AB, alpha, C, beta, 0.0f, result, -1, stream ) ;
}

// general matrix-multiply (GEMM): C = A * B * alpha + C * beta
void gemmCPU( const cv::Mat  A,
              const cv::Mat  B,
              const cv::Mat  C,
                    cv::Mat& result,
              const float    alpha,
              const float    beta   ) {
	
	// check input
	HAWAII_ERROR_CONDITIONAL( A.cols   != B.rows
	                       || A.rows   != C.rows
	                       || B.cols   != C.cols,
	                          "size mismatch" ) ;
	HAWAII_ERROR_CONDITIONAL( A.type() != CV_32FC1
	                       || B.type() != CV_32FC1
	                       || C.type() != CV_32FC1,
	                          "type must be single-channel float" ) ;
	// allocate result
	result.create( C.rows, C.cols, C.type() ) ;
	
	// wrap Eigen, partition to use multiple cores
	const auto BFull = cvMat2EigenMap( B ) ;
	const std::vector< cv::Range > ranges = hawaii::partitionize( C.rows, omp_get_max_threads() ) ;
	#pragma omp parallel for
	for( size_t part = 0 ; part < ranges.size() ; ++part ) {
		
		// extract and wrap partitions
		const auto      APart = cvMat2EigenMap(      A.rowRange( ranges[ part ] ) ) ;
		const auto      CPart = cvMat2EigenMap(      C.rowRange( ranges[ part ] ) ) ;
		      auto resultPart = cvMat2EigenMap( result.rowRange( ranges[ part ] ) ) ;
		
		// calculate GEMM
		resultPart = APart * BFull * alpha + CPart * beta ;
	}
}
void gemmGPU( const cv::gpu::GpuMat  A,
              const cv::gpu::GpuMat  B,
              const cv::gpu::GpuMat  C,
                    cv::gpu::GpuMat& result,
              const float            alpha,
              const float            beta,
                    cublasHandle_t   handle,
                    cv::gpu::Stream& stream ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( A.cols   != B.rows
	                       || A.rows   != C.rows
	                       || B.cols   != C.cols,
	                          "size mismatch" ) ;
	HAWAII_ERROR_CONDITIONAL( A.type() != CV_32FC1
	                       || B.type() != CV_32FC1
	                       || C.type() != CV_32FC1,
	                          "type must be single-channel float" ) ;
	// allocate result
	hawaii::GPU::memoryPool( result, C.rows, C.cols, C.type() ) ;
	
	// wrap OpenCV
	GEMMCUBLASWrapper( A, B, alpha, C, beta, result, handle, stream ) ;
}
