// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of demoHeterogeneousParallelism.
// 
// demoHeterogeneousParallelism is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// demoHeterogeneousParallelism is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
// even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
// 
// You should have received a copy of the GNU General Public License along with demoHeterogeneousParallelism. If not, 
// see <http://www.gnu.org/licenses/>.


// "libHawaii" processing steps to wrap existing code: dummy sample application
// ============================================================================

#pragma once

#include "hawaii/flow/item.h"
#include "hawaii/flow/step.h"
#include "hawaii/flow/stepGPU.h"
#include <unistd.h>

// dummy work item
class ItemDummy : public hawaii::flow::ItemTemplate< ItemDummy > {
	
	// require running number to be set
	public:
	ItemDummy( const uint64_t serialNumberArg,
	           const double   ratioArg        = 1.0 ) :
		ItemTemplate< ItemDummy >( serialNumberArg, ratioArg ) {
	}
	
	// split work item (required for partitioning only)
	public:
	virtual void splitImpl(       PtrVec&                     parts,
	                        const std::vector< double >&      ratios,
	                        const hawaii::flow::PartInfoBase& partInfo = hawaii::flow::PartInfoBase() ) override {
		// not needed
		(void)partInfo ;
		
		// create partitioned work items
		parts.clear() ;
		for( size_t idx = 0 ; idx < ratios.size() ; ++idx ) {
			parts.push_back( Ptr( new ItemDummy( this->serialNumber, ratios[ idx ] * this->ratio ) ) ) ;
		}
	}
	
	// re-combine work item (required for partitioning only)
	public:
	virtual void mergeImpl( const PtrVec&                     parts,
	                              Ptr&                        whole,
	                        const hawaii::flow::PartInfoBase& partInfo = hawaii::flow::PartInfoBase() ) override {
		// not needed
		(void)partInfo ;
		
		// sum up the ratios of all partitions
		double ratioWhole = 0.0 ;
		for( const auto& part : parts ) {
			ratioWhole += part->ratio ;
		}
		
		// create whole item: "ratioWhole" is not necessarily 1.0, e.g. with nested partitioning. 
		whole = Ptr( new ItemDummy( parts[ 0 ]->serialNumber, ratioWhole ) ) ;
	}
	
} ; // class "ItemDummy"

// source for dummy items
class SourceDummy : public hawaii::flow::StepTemplate< hawaii::flow::ItemBase,
                                                       ItemDummy              > {
	// default c'tor
	public:
	SourceDummy() :
		StepTemplate< hawaii::flow::ItemBase,
		              ItemDummy              >( "dummy source",          // name for printed evaluation
		                                        1,                       // use 1 CPU core
		                                        std::vector< int >(),    // don't use any GPUs
		                                        true,                    // source step
		                                        false                ) { // no sink step
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// copy serial number from source item
		dstPtr = DstPtr( new Dst( srcPtr->serialNumber ) ) ;
	}
} ;

// dummy first step on CPU
class FirstStepDummyCPU : public hawaii::flow::StepTemplate< ItemDummy > {
	
	// set name and processing time
	public:
	double m, c ;
	FirstStepDummyCPU() {
		this->name = "CPU dummy first step" ;
		this->m = 0.100 ;
		this->c = 0.000 ;
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// pass work item through
		dstPtr = srcPtr ;
		
		// wait according to item ratio
		usleep( 1000000.0 * ( srcPtr->ratio * this->m + this->c ) ) ;
	}
} ;

// dummy first step on GPU
class FirstStepDummyGPU : public hawaii::flow::StepTemplateGPU< ItemDummy > {
	
	// set name and processing time
	public:
	double m, c ;
	FirstStepDummyGPU() {
		this->name = "GPU dummy first step" ;
		this->m = 0.050 ;
		this->c = 0.000 ;
	}
	
	// inherited purely-virtual methods
	public:
	void uploadImpl( const SrcPtr&          srcPtr,
	                       SrcGPUPtr&       srcGPUPtr,
	                       cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		(void)stream ;
		srcGPUPtr = srcPtr ;
	}
	void computeImpl( const SrcGPUPtr&       srcGPUPtr,
	                        DstGPUPtr&       dstGPUPtr,
	                        cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		(void)stream ;
		dstGPUPtr = srcGPUPtr ;
		
		// wait according to item ratio
		usleep( 1000000.0 * ( srcGPUPtr->ratio * this->m + this->c ) ) ;
	}
	void downloadImpl( const DstGPUPtr&       dstGPUPtr,
	                         DstPtr&          dstPtr,
	                         cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		(void)stream ;
		dstPtr = dstGPUPtr ;
	}
} ;

// dummy second step on CPU
class SecondStepDummyCPU : public hawaii::flow::StepTemplate< ItemDummy > {
	
	// set name and processing time
	public:
	double m, c ;
	SecondStepDummyCPU() {
		this->name = "CPU dummy second step" ;
		this->m = 0.025 ;
		this->c = 0.000 ;
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// pass work item through
		dstPtr = srcPtr ;
		
		// wait according to item ratio
		usleep( 1000000.0 * ( srcPtr->ratio * this->m + this->c ) ) ;
	}
} ;

// dummy second step on GPU
class SecondStepDummyGPU : public hawaii::flow::StepTemplateGPU< ItemDummy > {
	
	// set name and processing time
	public:
	double m, c ;
	SecondStepDummyGPU() {
		this->name = "GPU dummy second step" ;
		this->m = 0.010 ;
		this->c = 0.000 ;
	}
	
	// inherited purely-virtual methods
	public:
	void uploadImpl( const SrcPtr&          srcPtr,
	                       SrcGPUPtr&       srcGPUPtr,
	                       cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		(void)stream ;
		srcGPUPtr = srcPtr ;
	}
	void computeImpl( const SrcGPUPtr&       srcGPUPtr,
	                        DstGPUPtr&       dstGPUPtr,
	                        cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		(void)stream ;
		dstGPUPtr = srcGPUPtr ;
		
		// wait according to item ratio
		usleep( 1000000.0 * ( srcGPUPtr->ratio * this->m + this->c ) ) ;
	}
	void downloadImpl( const DstGPUPtr&       dstGPUPtr,
	                         DstPtr&          dstPtr,
	                         cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		(void)stream ;
		dstPtr = dstGPUPtr ;
	}
} ;

// sink for dummy items
class SinkDummy : public hawaii::flow::StepTemplate< ItemDummy,
                                                     hawaii::flow::ItemBase > {
	// default c'tor
	public:
	SinkDummy() :
		StepTemplate< ItemDummy,
		              hawaii::flow::ItemBase >( "dummy sink",            // name for printed evaluation
		                                        1,                       // use 1 CPU core
		                                        std::vector< int >(),    // don't use any GPUs
		                                        false,                   // no source step
		                                        true                 ) { // sink step
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		// print message
		hawaii::cout << "finished dummy item #" << srcPtr->serialNumber << hawaii::endl ;
		
		// generate an output item for performance measurement even though this is a sink
		dstPtr = DstPtr( new Dst( srcPtr->serialNumber ) ) ;
	}
} ;
