// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of demoHeterogeneousParallelism.
// 
// demoHeterogeneousParallelism is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// demoHeterogeneousParallelism is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
// even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
// 
// You should have received a copy of the GNU General Public License along with demoHeterogeneousParallelism. If not, 
// see <http://www.gnu.org/licenses/>.


// place-holder for your code: feature matching sample application
// ===============================================================

#include "featureMatching.h"
#include "hawaii/GPU/pool.h"
#include "hawaii/common/error.h"

// detect and extract keypoints from one image
// developer note: Wrapped OpenCV code does not support streams for asynchronous operation.
void detectExtractCPU( const cv::Mat                      image,
	                          std::vector< cv::KeyPoint >& keypoints,
	                          cv::Mat&                     descriptors,
	                    const DetectExtractOptions         options     ) {
	
	// instantiate and use detector/extractor
	cv::ORB impl( options.nFeatures,
	              options.scaleFactor,
	              options.nLevels,
	              options.edgeThreshold,
	              options.firstLevel,
	              options.WTA_K,
	              options.scoreType,
	              options.patchSize ) ;
	impl( image, cv::Mat(), keypoints, descriptors ) ;
}
void detectExtractGPU( const cv::gpu::GpuMat              image,
	                          std::vector< cv::KeyPoint >& keypoints,
	                          cv::gpu::GpuMat&             descriptors,
	                    const DetectExtractOptions         options     ) {
	
	// instantiate and use detector/extractor
	// developer note: The CPU version always blurs the input image, the GPU version only if the flag below is set. We 
	//                 do so to make CPU and GPU descriptors compatible. 
	cv::gpu::ORB_GPU impl( options.nFeatures,
	                       options.scaleFactor,
	                       options.nLevels,
	                       options.edgeThreshold,
	                       options.firstLevel,
	                       options.WTA_K,
	                       options.scoreType,
	                       options.patchSize ) ;
	impl.blurForDescriptor = true ;
	impl( image, cv::gpu::GpuMat(), keypoints, descriptors ) ;
}

// find correspondences between features
void matchCPU( const cv::Mat  descriptorsCurr,
               const cv::Mat  descriptorsPrev,
                     cv::Mat& indices,
                     cv::Mat& distances ) {
	
	// instantiate and use matcher
	cv::BFMatcher impl( cv::NORM_HAMMING ) ;
	std::vector< std::vector< cv::DMatch > > matches ;
	impl.knnMatch( descriptorsCurr, descriptorsPrev,
	               matches,
	               1, cv::Mat(), false ) ;
	
	// convert matches to indices and distances
	indices.create(   1, matches.size(), CV_32SC1 ) ;
	distances.create( 1, matches.size(), CV_32FC1 ) ;
	for( int idxCurr = 0 ; idxCurr < indices.cols ; ++idxCurr ) {
		indices.at<   int   >( idxCurr ) = !matches[ idxCurr ].empty() ? matches[ idxCurr ][ 0 ].trainIdx : -1  ;
		distances.at< float >( idxCurr ) = !matches[ idxCurr ].empty() ? matches[ idxCurr ][ 0 ].distance : NAN ;
	}
}
void matchGPU( const cv::gpu::GpuMat& descriptorsCurr,
               const cv::gpu::GpuMat& descriptorsPrev,
                     cv::gpu::GpuMat& indices,
                     cv::gpu::GpuMat& distances,
                     cv::gpu::Stream& stream             ) {
	
	// allocate results
	hawaii::GPU::memoryPool( indices,   descriptorsCurr.rows, 1, CV_32SC1 ) ;
	hawaii::GPU::memoryPool( distances, descriptorsCurr.rows, 1, CV_32FC1 ) ;
	
	// instantiate and use matcher
	cv::gpu::BFMatcher_GPU impl( cv::NORM_HAMMING ) ;
	impl.matchSingle( descriptorsCurr, descriptorsPrev,
	                  indices, distances,
	                  cv::gpu::GpuMat(), stream ) ;
}

// visualize found matches
void drawMatches( const cv::Mat                      imageCurr,
                  const std::vector< cv::KeyPoint >& keypointsCurr,
                  const std::vector< cv::KeyPoint >& keypointsPrev,
                  const cv::Mat                      indices,
                        cv::Mat&                     visualization    ) {
	// check input
	HAWAII_ERROR_CONDITIONAL( imageCurr.depth() != CV_8U
	                       || ( imageCurr.channels() != 1
	                         && imageCurr.channels() != 3 ),
	                          "Input image must be unsigned 8-bit grayscale or BGR." ) ;
	HAWAII_ERROR_CONDITIONAL( indices.type() != CV_32SC1,
	                          "Index type must be signed 32-bit single-channel." ) ;
	HAWAII_ERROR_CONDITIONAL( keypointsCurr.size() != (size_t)indices.cols,
	                          "Number of current keypoints and indices must match." ) ;
	
	// ensure individual BGR output image
	if( imageCurr.channels() == 3 ) {
		imageCurr.copyTo( visualization ) ;
	} else {
		#if CV_MAJOR_VERSION == 3
			cv::cvtColor( imageCurr, visualization, cv::COLOR_GRAY2BGR ) ;
		#elif CV_MAJOR_VERSION == 2
			cv::cvtColor( imageCurr, visualization, CV_GRAY2BGR ) ;
		#else
			#error "unsupported version of OpenCV"
		#endif
	}
	
	// draw a line for each valid match
	for( int idxCurr = 0 ; idxCurr < indices.cols ; ++idxCurr ) {
		
		// valid match found?
		const int idxPrev = indices.at< int >( idxCurr ) ;
		if( idxPrev != -1 ) {
			
			// very simple filtering based on keypoints' relative position and angle
			const cv::KeyPoint keypointCurr = keypointsCurr[ idxCurr ],
			                   keypointPrev = keypointsPrev[ idxPrev ] ;
			const float distance = sqrtf( ( keypointCurr.pt.x - keypointPrev.pt.x ) * ( keypointCurr.pt.x - keypointPrev.pt.x )
			                            + ( keypointCurr.pt.y - keypointPrev.pt.y ) * ( keypointCurr.pt.y - keypointPrev.pt.y ) ),
			            angle    = fabsf( keypointCurr.angle
			                            - keypointPrev.angle ) ;
			if( distance < 75.0f && angle < 15.0f ) {
				
				// draw a red line
				cv::line( visualization,
				          keypointCurr.pt,
				          keypointPrev.pt,
				          cv::Scalar( 0, 0, 255 ) ) ;
			}
		}
	}
}
