// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of demoHeterogeneousParallelism.
// 
// demoHeterogeneousParallelism is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// demoHeterogeneousParallelism is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
// even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
// 
// You should have received a copy of the GNU General Public License along with demoHeterogeneousParallelism. If not, 
// see <http://www.gnu.org/licenses/>.


// "libHawaii" processing steps to wrap existing code: feature matching sample application
// =======================================================================================

#pragma once

#include "hawaii/flow/itemNamedImages.h"
#include "hawaii/flow/step.h"
#include "hawaii/flow/stepGPU.h"
#include "hawaii/GPU/autoMat.h"
#include "hawaii/common/partitionize.h"
#include "featureMatching.h"
#include "ompWrapper.h"

// work item for feature matching
class ItemFeatureMatching : public hawaii::flow::ItemTemplate< ItemFeatureMatching,
                                                               hawaii::flow::PartInfoImageOverlap > {
	// require running number to be set
	public:
	ItemFeatureMatching( const uint64_t serialNumberArg,
	                     const double   ratioArg        = 1.0 ) :
		ItemTemplate< ItemFeatureMatching, hawaii::flow::PartInfoImageOverlap >( serialNumberArg, ratioArg ) {
	}
	
	// input image
	public:
	hawaii::AutoMat imageCurr, imagePrev ;
	
	// keypoints, stored on CPU only (wrapped OpenCV code not asynchronous anyway)
	public:
	std::vector< cv::KeyPoint > keypointsCurr, keypointsPrev ;
	
	// descriptors
	hawaii::AutoMat descriptorsCurr, descriptorsPrev ;
	
	// matches: E.g. "matchIndices[ 5 ] == 3" means that the 5th current feature matches the 3rd previous one.
	hawaii::AutoMat matchIndices, matchDistances ;
	
	// split work item (required for partitioning only)
	public:
	virtual void splitImpl(       PtrVec&                             parts,
	                        const std::vector< double >&              ratios,
	                        const hawaii::flow::PartInfoImageOverlap& partInfo = hawaii::flow::PartInfoImageOverlap() ) override {
		
		// create output partitions
		parts.resize( ratios.size() ) ;
		for( size_t idx = 0 ; idx < parts.size() ; ++idx ) {
			parts[ idx ].reset( new ItemFeatureMatching( this->serialNumber, this->ratio * ratios[ idx ] ) ) ;
		}
		
		// split input image
		if( !this->imageCurr.empty() ) {
			const auto ranges = hawaii::partitionize( this->imageCurr.rows, ratios, partInfo.winHeight - 1 ) ;
			for( size_t idx = 0 ; idx < parts.size() ; ++idx ) {
				parts[ idx ]->imageCurr = this->imageCurr.rowRange( ranges[ idx ] ) ;
			}
		}
		
		// split current (but not previous) keypoints and descriptors
		if( !this->keypointsCurr.empty()
		 && !this->descriptorsCurr.empty() ) {
			const auto ranges = hawaii::partitionize( this->keypointsCurr.size(), ratios ) ;
			for( size_t idx = 0 ; idx < parts.size() ; ++idx ) {
				parts[ idx ]->keypointsCurr   = std::vector< cv::KeyPoint >( this->keypointsCurr.begin() + ranges[ idx ].start,
				                                                             this->keypointsCurr.begin() + ranges[ idx ].end   ) ;
				parts[ idx ]->descriptorsCurr = this->descriptorsCurr.rowRange( ranges[ idx ] ) ;
			}
		}
		
		// completely copy previous keypoints and descriptors
		if( !this->keypointsPrev.empty()
		 && !this->descriptorsPrev.empty() ) {
			for( size_t idx = 0 ; idx < parts.size() ; ++idx ) {
				parts[ idx ]->imagePrev       = this->imagePrev       ;
				parts[ idx ]->keypointsPrev   = this->keypointsPrev   ;
				parts[ idx ]->descriptorsPrev = this->descriptorsPrev ;
			}
		}
		
		// split current matching indices and distances
		if( !this->matchIndices.empty()
		 && !this->matchDistances.empty() ) {
			const auto ranges = hawaii::partitionize( this->matchIndices.cols, ratios ) ;
			for( size_t idx = 0 ; idx < parts.size() ; ++idx ) {
				parts[ idx ]->matchIndices   = this->matchIndices.colRange(   ranges[ idx ] ) ;
				parts[ idx ]->matchDistances = this->matchDistances.colRange( ranges[ idx ] ) ;
			}
		}
	}
	
	// re-combine work item (required for partitioning only)
	public:
	virtual void mergeImpl( const PtrVec&                             parts,
	                              Ptr&                                whole,
	                        const hawaii::flow::PartInfoImageOverlap& partInfo = hawaii::flow::PartInfoImageOverlap() ) override {
		
		// sum up the ratios and content sizes of all partitions, consider the overlap of the image partitions
		const int overlap = partInfo.winHeight - 1 ;
		double ratioWhole = 0.0 ;
		size_t idxKeypoints   = 0, numKeypoints   = 0       ;
		int    idxImageRows   = 0, numImageRows   = overlap,
		       idxDescriptors = 0, numDescriptors = 0,
		       idxIndices     = 0, numIndices     = 0,
		       idxDistances   = 0, numDistances   = 0       ;
		for( const auto& part : parts ) {
			ratioWhole     += part->ratio                    ;
			numImageRows   += part->imageCurr.rows - overlap ;
			numKeypoints   += part->keypointsCurr.size()     ;
			numDescriptors += part->descriptorsCurr.rows     ;
			numIndices     += part->matchIndices.cols        ;
			numDistances   += part->matchDistances.cols      ;
		}
		
		// allocate whole item 
		whole = Ptr( new ItemFeatureMatching( parts[ 0 ]->serialNumber, ratioWhole ) ) ;
		whole->imageCurr.writeCPU(       numImageRows,   parts[ 0 ]->imageCurr.cols,       parts[ 0 ]->imageCurr.type() ) ;
		whole->keypointsCurr.resize( numKeypoints ) ;
		whole->descriptorsCurr.writeCPU( numDescriptors, parts[ 0 ]->descriptorsCurr.cols, CV_8UC1                      ) ;
		whole->matchIndices.writeCPU(    1,              numIndices,                       CV_32SC1                     ) ;
		whole->matchDistances.writeCPU(  1,              numDistances,                     CV_32FC1                     ) ;
		
		// previous keypoints and descriptors have not been partitioned
		whole->imagePrev       = parts[ 0 ]->imagePrev       ;
		whole->keypointsPrev   = parts[ 0 ]->keypointsPrev   ;
		whole->descriptorsPrev = parts[ 0 ]->descriptorsPrev ;
		
		// explicitly copy the overlap of the first partition, because - being the first - it does not have a predecessor
		parts[ 0 ]->imageCurr.readCPU().rowRange( 0, overlap ).copyTo(
			whole->imageCurr.rowRange( idxImageRows, idxImageRows + overlap )
		) ;
		idxImageRows += overlap ;
		
		// copy partition content
		for( const auto& part : parts ) {
			
			// image except for overlap with previous partition
			const cv::Range rangeImage( overlap, part->imageCurr.rows ) ;
			part->imageCurr.readCPU().rowRange( rangeImage ).copyTo(
				whole->imageCurr.rowRange( idxImageRows, idxImageRows + rangeImage.size() )
			) ;
			idxImageRows += rangeImage.size() ;
			
			// keypoints
			if( numKeypoints > 0 ) {
				for( const auto& keypoint : part->keypointsCurr ) {
					whole->keypointsCurr[ idxKeypoints++ ] = keypoint ;
				}
			}
			
			// descriptors and matching results
			if( part->descriptorsCurr.rows > 0 ) { part->descriptorsCurr.readCPU().copyTo( whole->descriptorsCurr.rowRange( idxDescriptors, idxDescriptors + part->descriptorsCurr.rows ) ) ; idxDescriptors += part->descriptorsCurr.rows ; }
			if( part->matchIndices.cols    > 0 ) { part->matchIndices.readCPU().copyTo(    whole->matchIndices.colRange(    idxIndices,     idxIndices     + part->matchIndices.cols    ) ) ; idxIndices     += part->matchIndices.cols    ; }
			if( part->matchDistances.cols  > 0 ) { part->matchDistances.readCPU().copyTo(  whole->matchDistances.colRange(  idxDistances,   idxDistances   + part->matchDistances.cols  ) ) ; idxDistances   += part->matchDistances.cols  ; }
		}
	}
	
} ; // class "ItemFeatureMatching"

// feature detection and extraction on CPU
class DetectExtractCPU : public hawaii::flow::StepTemplate< hawaii::flow::ItemNamedImages,
                                                            ItemFeatureMatching           > {
	// options as members, set name and cores used
	public:
	DetectExtractOptions options ;
	DetectExtractCPU() {
		this->name = "CPU detect/extract" ;
		this->setCPUCores( omp_get_max_threads() ) ;
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// change work item type, but keep input image
		dstPtr = DstPtr( new ItemFeatureMatching( srcPtr->serialNumber,
		                                          srcPtr->ratio        ) ) ;
		dstPtr->imageCurr = srcPtr->images[ "orig" ] ;
		
		// wrap CPU function from "featureMatching.h", reduce number of features according to "ratio"
		DetectExtractOptions optionsCall = this->options ;
		optionsCall.nFeatures *= srcPtr->ratio ;
		detectExtractCPU( dstPtr->imageCurr,
		                  dstPtr->keypointsCurr,
		                  dstPtr->descriptorsCurr,
		                  optionsCall             ) ;
	}
} ;

// feature detection and extraction on GPU
// developer note: Wrapped OpenCV code does not support streams for asynchronous operation.
class DetectExtractGPU : public hawaii::flow::StepTemplate< hawaii::flow::ItemNamedImages,
                                                            ItemFeatureMatching           > {
	// options as members, set name
	public:
	DetectExtractOptions options ;
	DetectExtractGPU() {
		this->name = "GPU detect/extract" ;
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// change work item type, but keep input image
		dstPtr = DstPtr( new ItemFeatureMatching( srcPtr->serialNumber,
		                                          srcPtr->ratio        ) ) ;
		dstPtr->imageCurr = srcPtr->images[ "orig" ] ;
		
		// wrap GPU function from "featureMatching.h", reduce number of features according to "ratio"
		DetectExtractOptions optionsCall = this->options ;
		optionsCall.nFeatures *= srcPtr->ratio ;
		detectExtractGPU( dstPtr->imageCurr,
		                  dstPtr->keypointsCurr,
		                  dstPtr->descriptorsCurr,
		                  optionsCall             ) ;
	}
} ;

// buffer for saving descriptors from previous image
class BufferFeatureMatching : public hawaii::flow::StepTemplate< ItemFeatureMatching > {
	
	// c'tor with size of (square) matrices
	public:
	BufferFeatureMatching() {
		this->name = "feature matching buffer" ;
	}
	
	// buffered members from previous work item
	protected:
	mutable hawaii::AutoMat             imagePrev       ;
	mutable std::vector< cv::KeyPoint > keypointsPrev   ;
	mutable hawaii::AutoMat             descriptorsPrev ;
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// pass work item through
		dstPtr = srcPtr ;
		
		// add buffered data to current item, then update buffers
		dstPtr->imagePrev       = this->imagePrev         ;
		dstPtr->keypointsPrev   = this->keypointsPrev     ;
		dstPtr->descriptorsPrev = this->descriptorsPrev   ;
		this->imagePrev         = srcPtr->imageCurr       ;
		this->keypointsPrev     = srcPtr->keypointsCurr   ;
		this->descriptorsPrev   = srcPtr->descriptorsCurr ;
	}
} ;

// feature matching on CPU
class MatchCPU : public hawaii::flow::StepTemplate< ItemFeatureMatching > {
	
	// set name and cores used
	public:
	MatchCPU() {
		this->name = "CPU match" ;
		this->setCPUCores( omp_get_max_threads() ) ;
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// pass work item through
		dstPtr = srcPtr ;
		
		// wrap CPU function from "featureMatching.h"
		if( HAWAII_LIKELY( !srcPtr->descriptorsPrev.empty() ) ) {
			matchCPU( srcPtr->descriptorsCurr,
			          srcPtr->descriptorsPrev,
			          dstPtr->matchIndices,
			          dstPtr->matchDistances  ) ;
		}
	}
} ;

// feature matching on GPU
class MatchGPU : public hawaii::flow::StepTemplateGPU< ItemFeatureMatching > {
	
	// set name
	public:
	MatchGPU() {
		this->name = "GPU match" ;
	}
	
	// inherited purely-virtual methods
	public:
	void uploadImpl( const SrcPtr&          srcPtr,
	                       SrcGPUPtr&       srcGPUPtr,
	                       cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		srcGPUPtr = srcPtr ;
		
		// upload inputs to GPU
		srcGPUPtr->descriptorsCurr.readGPU( stream ) ;
		if( HAWAII_LIKELY( !srcGPUPtr->descriptorsPrev.empty() ) ) {
			srcGPUPtr->descriptorsPrev.readGPU( stream ) ;
		}
	}
	void computeImpl( const SrcGPUPtr&       srcGPUPtr,
	                        DstGPUPtr&       dstGPUPtr,
	                        cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstGPUPtr = srcGPUPtr ;
		
		// wrap GPU function from "featureMatching.h"
		if( HAWAII_LIKELY( !srcGPUPtr->descriptorsPrev.empty() ) ) {
			matchGPU( srcGPUPtr->descriptorsCurr,
			          srcGPUPtr->descriptorsPrev,
			          dstGPUPtr->matchIndices,
			          dstGPUPtr->matchDistances,
			          stream                     ) ;
		}
	}
	void downloadImpl( const DstGPUPtr&       dstGPUPtr,
	                         DstPtr&          dstPtr,
	                         cv::gpu::Stream& stream    ) override {
		
		// pass work item through
		dstPtr = dstGPUPtr ;
		
		// download results from GPU
		if( HAWAII_LIKELY( !dstPtr->matchIndices.empty() ) ) {
			dstPtr->matchIndices.readCPU(   stream ) ;
		}
		if( HAWAII_LIKELY( !dstPtr->matchDistances.empty() ) ) {
			dstPtr->matchDistances.readCPU( stream ) ;
		}
	}
} ;

// sink to draw the found matches
class SinkFeatureMatching : public hawaii::flow::StepTemplate< ItemFeatureMatching,
                                                               hawaii::flow::ItemBase > {
	// c'tor to create a sink
	public:
	SinkFeatureMatching() :
		StepTemplate< ItemFeatureMatching,
		              hawaii::flow::ItemBase >( "feature matching visualization",
		                                        1, std::vector< int >(),
		                                        false, true                      ) {
	}
	
	// inherited purely-virtual method
	public:
	void processImpl( const SrcPtr& srcPtr,
	                        DstPtr& dstPtr ) override {
		
		// render and display matches as lines
		cv::Mat visualization ;
		if( HAWAII_LIKELY( !srcPtr->matchIndices.empty() ) ) {
			cv::Mat blend ;
			cv::addWeighted( srcPtr->imageCurr, 0.5,
			                 srcPtr->imagePrev, 0.5,
			                 0.0,
			                 blend, -1              ) ;
			drawMatches( blend,
			             srcPtr->keypointsCurr,
			             srcPtr->keypointsPrev,
			             srcPtr->matchIndices,
			             visualization         ) ;
		} else {
			visualization = srcPtr->imageCurr ;
		}
		HAWAII_IMSHOW( visualization ) ;
		cv::waitKey( 1 ) ;
		
		// generate an output item for performance measurement even though this is a sink
		dstPtr = DstPtr( new Dst( srcPtr->serialNumber ) ) ;
	}
} ;
