// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// queues to buffer work items
// ===========================

#pragma once

#include "hawaii/flow/iface.h"
#include "hawaii/common/helpers.h"
#include <opencv2/core/core.hpp>
#include "hawaii/GPU/nvccFixHeaderAbove.h"
#include <boost/smart_ptr/shared_ptr.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/locks.hpp>
#include <boost/thread/condition.hpp>
#include <boost/circular_buffer.hpp>
#include "hawaii/GPU/nvccFixHeaderBelow.h"
#include <queue>

namespace hawaii {
namespace flow {

// factory function for any of the queues below
enum QueueType {
	queueTypeFIFO,
	queueTypeFIFOOverflow,
	queueTypePriority,
	queueTypePriorityOverflow,
	queueTypeOrdered,
	queueTypeOrderedOverflow
} ;
class QueueBase ;
QueueBase* queueFactory( const QueueType queueTypeArg = queueTypeFIFO,
                         const size_t    capacityArg  = 1             ) ;

// common part of all queues
// developer note: Methods without suffix "Unlocked" are public and thread-safe, the ones with the suffix are protected 
//                 and do not perform any locking. Making that difference allows "this->mutex" to be a "boost::mutex" 
//                 instead of a "boost::recursive_mutex" and consequently "this->{pull,push}able" to be a "boost::
//                 condition_variable" instead of a "boost::condition_variable_any", which would have been slower.
class QueueBase : public IfacePush,
                  public IfacePull {
	
	// pointer for classes wrapping queues
	public:
	typedef boost::shared_ptr< hawaii::flow::QueueBase > Ptr    ;
	typedef std::vector< Ptr >                           PtrVec ;
	
	// efficient thread-safety: wait until other threads are done, but no longer than that
	protected:
	typedef boost::mutex              Mutex ; mutable Mutex mutex              ;
	typedef boost::condition_variable Cond  ;         Cond  pullable, pushable ;
	
	// actual implementation
	protected:
	virtual bool pushTryUnlocked( const ItemBase::Ptr& itemPtr ) = 0 ;
	virtual bool pullTryUnlocked(       ItemBase::Ptr& itemPtr ) = 0 ;
	
	// inherited push/pull interface, override only if necessary
	protected:
	virtual void pushImpl(    const ItemBase::Ptr& itemPtr ) override ;
	virtual bool pushTryImpl( const ItemBase::Ptr& itemPtr ) override ;
	virtual void pullImpl(          ItemBase::Ptr& itemPtr ) override ;
	virtual bool pullTryImpl(       ItemBase::Ptr& itemPtr ) override ;
	
	// get additional information
	protected:
	virtual size_t capacityUnlocked() const = 0 ; // maximum possible number of contained items
	virtual size_t sizeUnlocked()     const = 0 ; // current number of contained items
	        size_t freeUnlocked()     const { return this->capacityUnlocked() -  this->sizeUnlocked() ; }
	        bool   fullUnlocked()     const { return this->capacityUnlocked() <= this->sizeUnlocked() ; }
	        bool   emptyUnlocked()    const { return 0                        >= this->sizeUnlocked() ; }
	public:
	size_t capacity() const { boost::lock_guard< Mutex >( this->mutex ) ; return this->capacityUnlocked() ; }
	size_t size()     const { boost::lock_guard< Mutex >( this->mutex ) ; return this->sizeUnlocked()     ; }
	size_t free()     const { boost::lock_guard< Mutex >( this->mutex ) ; return this->freeUnlocked()     ; }
	bool   full()     const { boost::lock_guard< Mutex >( this->mutex ) ; return this->fullUnlocked()     ; }
	bool   empty()    const { boost::lock_guard< Mutex >( this->mutex ) ; return this->emptyUnlocked()    ; }
	
} ; // abstract class "QueueBase"

// FIFO queue with overflow: Items can always be pushed in but - if capacity is exceeded - the earliest ones are lost.
class QueueFIFOOverflow : public QueueBase {
	
	// simply wrap "boost::circular_buffer", which behaves exactly as described above
	protected:
	boost::circular_buffer< ItemBase::Ptr > impl ;
	virtual bool pushTryUnlocked( const ItemBase::Ptr& itemPtr ) override ;
	virtual bool pullTryUnlocked(       ItemBase::Ptr& itemPtr ) override ;
	
	// get additional information
	public:
	size_t capacityUnlocked() const override { return this->impl.capacity() ; }
	size_t sizeUnlocked()     const override { return this->impl.size()     ; }
	
	// c'tor with optional capacity limit
	public:
	QueueFIFOOverflow( const size_t capacityArg = 2 ) : impl( capacityArg ) {}
} ;

// FIFO queue without overflow: If capacity is reached, pushing new items blocks or fails.
class QueueFIFO : public QueueFIFOOverflow {
	
	// change push interface to from overflow to rejecting
	protected:
	virtual bool pushTryUnlocked( const ItemBase::Ptr& itemPtr ) override ;
	
	// c'tor with optional capacity limit
	public:
	QueueFIFO( const size_t capacityArg = 2 ) : QueueFIFOOverflow( capacityArg ) {}
} ;

// allow items with a lower serial number to "overtake" others, with overflow like "QueueFIFOOverflow"
class QueuePriorityOverflow : public QueueBase {
	
	// work item sorter for use within standard library containers and functions
	// developer note: The standard library's sorting is undefined for equal priorities, so no matter if ">" or ">=" 
	//                 is used, this will mix FIFO and LIFO behavior. Therefore, definitely set an increasing serial 
	//                 number for your derived work items.
	protected:
	struct ItemPtrSorter {
		bool operator ()( const ItemBase::Ptr& left,
		                  const ItemBase::Ptr& right ) {
			return left->serialNumber > right->serialNumber ;
		}
	} ;
	
	// "std::priority_queue" does the sorting, but not the overflow.
	protected:
	std::priority_queue< ItemBase::Ptr, ItemBase::PtrVec, ItemPtrSorter > impl ;
	const size_t implCapacity ;
	virtual bool pushTryUnlocked( const ItemBase::Ptr& itemPtr ) override ;
	virtual bool pullTryUnlocked(       ItemBase::Ptr& itemPtr ) override ;
	
	// get additional information
	public:
	size_t capacityUnlocked() const override { return this->implCapacity ; }
	size_t sizeUnlocked()     const override { return this->impl.size()  ; }
	
	// c'tor with optional capacity limit
	public:
	QueuePriorityOverflow( const size_t capacityArg = 2 ) : implCapacity( capacityArg ) {}
} ;

// allow items with a lower serial number to "overtake" others, without overflow like "QueueFIFO"
class QueuePriority : public QueuePriorityOverflow {
	
	// change push interface to from overflow to rejecting
	protected:
	virtual bool pushTryUnlocked( const ItemBase::Ptr& itemPtr ) override ;
	
	// c'tor with optional capacity limit
	public:
	QueuePriority( const size_t capacityArg = 2 ) : QueuePriorityOverflow( capacityArg ) {}
} ;

// offer items to be pulled ordered by serial number
// user note: Gaps w.r.t. serial numbers may occur for two reasons:
//            - The oldest item may be dropped due to overflow (see above).
//            - If e.g. item #4 has already been pulled before item #2 is pushed, #2 will be dropped on arrival.
class QueueOrderedOverflow : public QueuePriorityOverflow {
	
	// keep track of last pulled item
	protected:
	uint64_t serialNumberPulled ;
	
	// change push interface to drop "out-dated" items on arrival
	protected:
	virtual bool pushTryUnlocked( const ItemBase::Ptr& itemPtr ) override ;
	
	// change pull interface to update last pulled serial number
	protected:
	virtual bool pullTryUnlocked( ItemBase::Ptr& itemPtr ) override ;
	
	// c'tor with optional capacity limit
	public:
	QueueOrderedOverflow( const size_t capacityArg = 2 ) : QueuePriorityOverflow( capacityArg ), serialNumberPulled( 0 ) {}
} ;

// offer items to be pulled with consecutive serial numbers
// user note: In contrast to "QueueOrdered", gaps cannot occur, but pushing can block or fail.
class QueueOrdered : public QueuePriority {
	
	// keep track of next item to be pulled
	protected:
	uint64_t serialNumberNext ;
	
	// change push interface to reject items with too high serial numbers: If e.g. #5 must be pulled next and capacity 
	// is 4, only accept #5-8 to be pushed; #9 must wait until #5 is out again.
	protected:
	virtual bool pushTryUnlocked( const ItemBase::Ptr& itemPtr ) override ;
	
	// change pull interface to block or fail if the next consecutive item is not here yet
	protected:
	virtual bool pullTryUnlocked( ItemBase::Ptr& itemPtr ) override ;
	
	// also override public pull interface to notify all instead of just one pushing thread, because the one's item may 
	// be rejected 
	public:
	virtual void pullImpl(    ItemBase::Ptr& itemPtr ) override ;
	virtual bool pullTryImpl( ItemBase::Ptr& itemPtr ) override ;
	
	// c'tor with optional capacity limit
	public:
	QueueOrdered( const size_t capacityArg = 2 ) : QueuePriority( capacityArg ), serialNumberNext( 0 ) {}
} ;

} } // namespaces "hawaii::flow"
