// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// series of processing steps to be executed sequentially
// ======================================================

#pragma once

#include "hawaii/GPU/nvccProhibit.h"
#include "hawaii/flow/step.h"
#include <boost/serialization/shared_ptr.hpp>

namespace hawaii {
namespace flow {

// combine several processing steps into one
// user note #1: This is not a pipeline - steps run one after another.
// user note #2: GPU latency hiding via overlapping memory copies and kernels as implemented in "Step{Base,Template}GPU" 
//               will be disabled when adding them to a series.
class Series : public StepBase {
	
	// c'tor to directly add any number of processing steps and other arguments for which "Series::add()" is defined
	public:
	template< typename ... Args >
	Series( Args&& ... args ) ;
	
	// add one processing step at a time
	public:
	template< typename ... Args >
	void add( hawaii::flow::StepBase& stepNext,
	          Args&& ...              argsRest ) ;
	
	// allow the variadic template to set the class's name and to terminate
	protected:
	template< typename ... Args >
	void add( const std::string nameArg,
	                Args&& ...  argsRest ) ;
	void add() {}
	
	// reset internal state
	public:
	virtual void reset() ;
	
	// contain pointers to steps instead of copies
	protected:
	StepBase::PtrVec steps ;
	
	// subsequently call all steps
	public:
	virtual void process( const ItemBase::Ptr& srcPtr,
	                            ItemBase::Ptr& dstPtr ) override ;
} ; // class "Series"

// "Series": combine several processing steps into one
// ---------------------------------------------------

// c'tor to directly add any number of processing steps and other arguments for which "Series::add()" is defined
template< typename ... Args >
inline Series::Series( Args&& ... args ) :
	StepBase( "series of processing steps (unnamed)",
	          1, std::vector< int >(),
	          true, true                             ) {
	this->add( std::forward< Args >( args ) ... ) ;
}

// add one processing step at a time
template< typename ... Args >
inline void Series::add( hawaii::flow::StepBase& stepNext,
                         Args&& ...              argsRest ) {
	
	// act as a source/sink if first/last step is one respectively
	if( this->steps.empty() ) {
		this->source = stepNext.isSource() ;
	}
	this->sink = stepNext.isSink() ;
	
	// update hardware to maximum number of CPU cores and union of GPUs used
	this->CPUCores = std::max( this->CPUCores, stepNext.getCPUCores() ) ;
	this->GPUBits |= stepNext.getGPUBits() ;
	
	// wrap processing step and continue with next
	this->steps.push_back( StepBase::Ptr( &stepNext, boost::serialization::null_deleter() ) ) ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}

// allow the variadic template to set the class's name
template< typename ... Args >
inline void Series::add( const std::string nameArg,
                               Args&& ...  argsRest ) {
	this->name = nameArg ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}

} } // namespaces "hawaii::flow"
