// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// base classes for GPU processing steps to overlap computations and memory copies
// ===============================================================================

#pragma once

#include "hawaii/flow/step.h"
#include "hawaii/GPU/openCV.h"

namespace hawaii {
namespace flow {

// (really) base GPU processing step: see "StepTemplateGPU" below before deriving from this
class StepBaseGPU : public StepBase {
	
	// Processing must be split into three stages: Use the given stream for anything related to CUDA. Avoid implicit 
	//                                             synchronization - e.g. use "hawaii::GPU::memoryPool" instead of 
	//                                             "cudaMalloc*" or "cv::gpu::{GpuMat::CudaMem}::create()".
	public:
	// enqueue upload of input data to the provided stream
	virtual void upload(   const hawaii::flow::ItemBase::Ptr& srcBasePtr,
	                             hawaii::flow::ItemBase::Ptr& srcBaseGPUPtr,
	                             cv::gpu::Stream&             stream        ) = 0 ;
	// enqueue kernel computations to the provided stream
	virtual void compute(  const hawaii::flow::ItemBase::Ptr& srcBaseGPUPtr,
	                             hawaii::flow::ItemBase::Ptr& dstBaseGPUPtr,
	                             cv::gpu::Stream&             stream        ) = 0 ;
	// enqueue download of output data to the provided stream
	virtual void download( const hawaii::flow::ItemBase::Ptr& dstBaseGPUPtr,
	                             hawaii::flow::ItemBase::Ptr& dstBasePtr,
	                             cv::gpu::Stream&             stream        ) = 0 ;
	
	// processing function to wrap the above
	// user note: This does not do any overlapping and will not be called by "run()" or "threadFunc()" below. It will 
	//            however be used e.g. when wrapping this processing step in a "hawaii::flow::Series".
	public:
	virtual void process( const ItemBase::Ptr& srcBasePtr,
	                            ItemBase::Ptr& dstBasePtr ) override final ;
	
	// c'tor with one specific GPU to run on
	// user note: A GPU step can be neither a source nor a sink.
	public:
	StepBaseGPU( const std::string nameArg = "GPU processing step (unnamed)",
	             const int         GPUArg  = 0                               ) ;
	
	// reset internal state
	public:
	virtual void reset() override {
		this->counterItemsNoOverlap = 0 ;
		StepBase::reset() ;
	}
	
	// print evaluation
	public:
	virtual void eval( size_t indentationName = 0,
	                   size_t indentationData = 0 ) const override ;
	
	// run this processing step
	public:
	virtual void run() override ;
	
	// thread function to implement overlapping computations and memory copies
	// user note #1: The overriding implementation will only be used if this step's selected GPU actually has 
	//               own memory and it has not been disabled manually.
	// user note #2: Due to this overlapping, a GPU step usually contains not just one, but up to three pairs of 
	//               source/destination work items at any given time.
	public:
	bool overlapKernelsAndCopies ;
	protected:
	uint64_t counterItemsNoOverlap ;
	void threadFuncGPU() ;
	
} ; // class "StepBaseGPU"

// base GPU processing step to directly use your work item type(s) instead of casting
template< typename ItemSrc,
          typename ItemDst    = ItemSrc,
          typename ItemSrcGPU = ItemSrc,
          typename ItemDstGPU = ItemDst >
class StepTemplateGPU : public StepBaseGPU {
	
	// shorten code
	public:
	typedef ItemSrc    Src    ; typedef boost::shared_ptr< Src    > SrcPtr    ;
	typedef ItemDst    Dst    ; typedef boost::shared_ptr< Dst    > DstPtr    ;
	typedef ItemSrcGPU SrcGPU ; typedef boost::shared_ptr< SrcGPU > SrcGPUPtr ;
	typedef ItemDstGPU DstGPU ; typedef boost::shared_ptr< DstGPU > DstGPUPtr ;
	
	// require purely virtual processing functions with actually used instead of base types
	// user note: see above
	public:
	// enqueue upload of input data to the provided stream
	virtual void uploadImpl( const SrcPtr&          srcPtr,
	                               SrcGPUPtr&       srcGPUPtr,
	                               cv::gpu::Stream& stream    ) = 0 ;
	virtual void upload( const ItemBase::Ptr&   srcBasePtr,
	                           ItemBase::Ptr&   srcBaseGPUPtr,
	                           cv::gpu::Stream& stream        ) override final ;
	// enqueue kernel computations to the provided stream
	virtual void computeImpl( const SrcGPUPtr&       srcGPUPtr,
	                                DstGPUPtr&       dstGPUPtr,
	                                cv::gpu::Stream& stream    ) = 0 ;
	virtual void compute( const ItemBase::Ptr&   srcBaseGPUPtr,
	                            ItemBase::Ptr&   dstBaseGPUPtr,
	                            cv::gpu::Stream& stream        ) override final ;
	// enqueue download of output data to the provided stream
	virtual void downloadImpl( const DstGPUPtr&       dstGPUPtr,
	                                 DstPtr&          dstPtr,
	                                 cv::gpu::Stream& stream    ) = 0 ;
	virtual void download( const ItemBase::Ptr&   dstBaseGPUPtr,
	                             ItemBase::Ptr&   dstBasePtr,
	                             cv::gpu::Stream& stream        ) override final ;
	
	// for convenience, also wrap "process()" function with actually used types instead of base types
	public:
	virtual void processImpl( const SrcPtr& srcPtr,
	                                DstPtr& dstPtr ) final ;
	
	// c'tor with one specific GPU to run on
	// developer note: A GPU step can be neither a source nor a sink.
	public:
	StepTemplateGPU( const std::string nameArg = "GPU processing step (unnamed)",
	                 const int         GPUArg  = 0                               ) :
		StepBaseGPU( nameArg,
		             GPUArg  ) {
	}
	
} ; // class template "StepTemplateGPU"

// "StepTemplateGPU": base GPU processing step to directly use your work item type(s) instead of casting
// -----------------------------------------------------------------------------------------------------

// wrap the actual-types processing functions
template< typename ItemSrc, typename ItemDst, typename ItemSrcGPU, typename ItemDstGPU >
void StepTemplateGPU< ItemSrc, ItemDst, ItemSrcGPU, ItemDstGPU >::upload( const ItemBase::Ptr&   srcBasePtr,
                                                                                ItemBase::Ptr&   srcBaseGPUPtr,
                                                                                cv::gpu::Stream& stream        ) {
	// try to cast to actual work items
	const SrcPtr    srcPtr    = boost::dynamic_pointer_cast< Src    >( srcBasePtr    ) ;
			SrcGPUPtr srcGPUPtr = boost::dynamic_pointer_cast< SrcGPU >( srcBaseGPUPtr ) ;
	
	// only allow a source step to run without input
	HAWAII_ERROR_CONDITIONAL( !this->source && !srcBasePtr, "obtained non-existent source work item"  ) ;
	HAWAII_ERROR_CONDITIONAL( !this->source && !srcPtr,     "obtained source work item of wrong type" ) ;
	
	// do the processing, "srcGPUPtr" may have to be allocated there so update base pointer afterwards
	this->activateGPU() ;
	this->uploadImpl( srcPtr, srcGPUPtr, stream ) ;
	srcBaseGPUPtr = boost::static_pointer_cast< hawaii::flow::ItemBase >( srcGPUPtr ) ;
	HAWAII_ERROR_CONDITIONAL( !srcBaseGPUPtr, "produced non-existent destination work item" ) ;
}
template< typename ItemSrc, typename ItemDst, typename ItemSrcGPU, typename ItemDstGPU >
void StepTemplateGPU< ItemSrc, ItemDst, ItemSrcGPU, ItemDstGPU >::compute( const ItemBase::Ptr&   srcBaseGPUPtr,
                                                                                 ItemBase::Ptr&   dstBaseGPUPtr,
                                                                                 cv::gpu::Stream& stream        ) {
	// try to cast to actual work items
	const SrcGPUPtr srcGPUPtr = boost::dynamic_pointer_cast< SrcGPU >( srcBaseGPUPtr ) ;
			DstGPUPtr dstGPUPtr = boost::dynamic_pointer_cast< DstGPU >( dstBaseGPUPtr ) ;
	
	// only allow a source step to run without input
	HAWAII_ERROR_CONDITIONAL( !this->source && !srcBaseGPUPtr, "obtained non-existent source work item"  ) ;
	HAWAII_ERROR_CONDITIONAL( !this->source && !srcGPUPtr,     "obtained source work item of wrong type" ) ;
	
	// do the processing, "dstGPUPtr" may have to be allocated there so update base pointer afterwards
	this->activateGPU() ;
	this->computeImpl( srcGPUPtr, dstGPUPtr, stream ) ;
	dstBaseGPUPtr = boost::static_pointer_cast< hawaii::flow::ItemBase >( dstGPUPtr ) ;
	HAWAII_ERROR_CONDITIONAL( !this->sink && !dstBaseGPUPtr, "produced non-existent destination work item" ) ;
}
template< typename ItemSrc, typename ItemDst, typename ItemSrcGPU, typename ItemDstGPU >
void StepTemplateGPU< ItemSrc, ItemDst, ItemSrcGPU, ItemDstGPU >::download( const ItemBase::Ptr&   dstBaseGPUPtr,
                                                                                  ItemBase::Ptr&   dstBasePtr,
                                                                                  cv::gpu::Stream& stream        ) {
	// try to cast to actual work items
	const DstGPUPtr dstGPUPtr = boost::dynamic_pointer_cast< DstGPU >( dstBaseGPUPtr ) ;
			DstPtr    dstPtr    = boost::dynamic_pointer_cast< Dst    >( dstBasePtr    ) ;
	
	// only allow a source step to run without input
	HAWAII_ERROR_CONDITIONAL( !this->source && !dstBaseGPUPtr, "obtained non-existent source work item"  ) ;
	HAWAII_ERROR_CONDITIONAL( !this->source && !dstGPUPtr,     "obtained source work item of wrong type" ) ;
	
	// do the processing, "dstPtr" may have to be allocated there so update base pointer afterwards
	this->activateGPU() ;
	this->downloadImpl( dstGPUPtr, dstPtr, stream ) ;
	dstBasePtr = boost::static_pointer_cast< hawaii::flow::ItemBase >( dstPtr ) ;
	HAWAII_ERROR_CONDITIONAL( !this->sink && !dstBasePtr, "produced non-existent destination work item" ) ;
}

// for convenience, also wrap "process()" function with actually used types instead of base types
template< typename ItemSrc, typename ItemDst, typename ItemSrcGPU, typename ItemDstGPU >
void StepTemplateGPU< ItemSrc, ItemDst, ItemSrcGPU, ItemDstGPU >::processImpl( const SrcPtr& srcPtr,
                                                                                     DstPtr& dstPtr ) {
	// cast to base work item
	const ItemBase::Ptr srcBasePtr = boost::static_pointer_cast< ItemBase >( srcPtr ) ;
			ItemBase::Ptr dstBasePtr = boost::static_pointer_cast< ItemBase >( dstPtr ) ;
	
	// only allow a source step to run without input
	HAWAII_ERROR_CONDITIONAL( !srcBasePtr, "obtained non-existent source work item" ) ;
	
	// do the processing, "dstPtr" may have to be allocated there so update base pointer afterwards
	this->activateGPU() ;
	StepBaseGPU::process( srcBasePtr, dstBasePtr ) ;
	dstPtr = boost::dynamic_pointer_cast< Dst >( dstBasePtr ) ;
	HAWAII_ERROR_CONDITIONAL( !this->sink && !dstBasePtr, "produced non-existent destination work item"  ) ;
	HAWAII_ERROR_CONDITIONAL( !this->sink && !dstPtr,     "produced destination work item of wrong type" ) ;
}

} } // namespaces "hawaii::flow"
