// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// processing step to save images to disk
// ======================================

#include "hawaii/IO/save.h"
#include "hawaii/common/error.h"
#include <opencv2/highgui/highgui.hpp>
#include <cstdio>

namespace hawaii {
namespace IO {

// save image with filename generated from pattern, label and index
void saveIndexed( const cv::Mat     image,
                  const std::string pattern,
                  const std::string label,
                  const uint64_t    index   ) {
	
	// generate file name from pattern and index
	constexpr size_t bufferSize = 4096 ;
	char filename[ bufferSize ] ;
	HAWAII_ERROR_CONDITIONAL( 0 > snprintf( filename, bufferSize,
	                                        pattern.c_str(),
	                                        label.c_str(),
	                                        index                ),
	                          "\"snprintf\" failed, increase \"bufferSize\"" ) ;
	const std::string file( filename ) ;
	
	// save if possible
	const int  type = image.type() ;
	const bool png  = file.length() >= 5 && file.compare( file.length() - 4, 4, ".png" ) ;
	HAWAII_ERROR_CONDITIONAL( type != CV_8UC1 && type != CV_16UC1
	                       && type != CV_8UC3 && type != CV_16UC3
	                       && type != CV_8UC4 && type != CV_16UC4,
	                          "image type not supported" ) ;
	HAWAII_ERROR_CONDITIONAL( !png && ( type == CV_8UC4 || type == CV_16UC1 || type == CV_16UC3 || type == CV_16UC4 ),
	                          "image type only supported by \".png\" format" ) ;
	HAWAII_ERROR_CONDITIONAL( !cv::imwrite( file, image ),
	                          "failed to save even though format should be supported" ) ;
}

// implement inherited purely-virtual method to save all images within the item
void StepSave::processImpl( const SrcPtr& srcPtr,
                                  DstPtr& dstPtr ) {
	
	// generate an output item for performance measurement even though this is a sink
	dstPtr = DstPtr( new Dst( srcPtr->serialNumber ) ) ;
	
	// save all the images
	for( const auto& image : srcPtr->images ) {
		
		// save image
		hawaii::IO::saveIndexed( image.second, this->filePattern, image.first, this->indexStart + srcPtr->serialNumber ) ;
	}
}

} } // namespaces "hawaii::IO"
