// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// demand-based allocation of work items to available processing units
// ===================================================================

#include "hawaii/flow/demandAlloc.h"

namespace hawaii {
namespace flow {

// "DemandAlloc": wrapper around identical functionality on different devices
// --------------------------------------------------------------------------

// reset internal state
void DemandAlloc::reset() {
	this->pullFinalItem = false ;
	for( auto& flow : this->flows ) { flow->reset() ; }
	FlowBase::reset() ;
}

// run this demand-based allocator, wait for it to finish or directly stop it
void DemandAlloc::run()  { this->stop() ;
                           this->reset() ;
                           for( FlowBase::Ptr& flow : this->flows ) { flow->run()  ; } }
void DemandAlloc::wait() { for( FlowBase::Ptr& flow : this->flows ) { flow->wait() ; } }
void DemandAlloc::stop() { for( FlowBase::Ptr& flow : this->flows ) { flow->stop() ; } }

// estimate power consumption
double DemandAlloc::power() const {
	double ret = 0.0 ;
	for( const auto& flow : this->flows ) {
		ret += flow->power() ;
	}
	return ret ;
}

// update contained flows' throughput limits to optimize power efficiency
void DemandAlloc::updateThroughputLimits() {
	
	// maximize performance in case throughput is not limited or required information is not yet available
	bool unlimited = this->throughputLimitDemandAlloc == INFINITY
	              && this->flowPullPtr                == nullptr  ;
	for( auto& flow : this->flows ) {
		unlimited |= isnan( flow->throughput() ) ;
		unlimited |= isnan( flow->power()      ) ;
	}
	if( unlimited ) {
		for( auto& flow : this->flows ) {
			flow->setThroughputLimit( INFINITY ) ;
		}
		return ;
	}
	
	// sort flow components by power efficiency
	std::vector< size_t > indices ; indices.reserve( this->flows.size() ) ;
	for( size_t index = 0 ; index < this->flows.size() ; ++index ) { indices.push_back( index ) ; }
	std::sort( indices.begin(), indices.end(),
	           [ this ]( size_t index1, size_t index2 ) {
		           return ( this->flows[ index1 ]->throughput() / this->flows[ index1 ]->power()
		                  > this->flows[ index2 ]->throughput() / this->flows[ index2 ]->power() ) ;
	           } ) ;
	
	// impose throughput limits on less efficient devices
	double throughputTarget = this->getThroughputLimit() ;
	if( this->flowPullPtr != nullptr
	 && this->flowPullPtr->waitedOnPush() < 0.1 ) {
		throughputTarget = std::min( throughputTarget,
		                             this->flowPullPtr->throughput() * 1.1 ) ;
	}
	double throughputRemaining = throughputTarget ;
	for( const auto index : indices ) {
		
		// full throttle as long as enough throughput left
		if( throughputRemaining > this->flows[ index ]->throughputUnconstrained() ) {
			this->flows[ index ]->setThroughputLimit( INFINITY ) ;
			throughputRemaining -= this->flows[ index ]->throughputUnconstrained() ; }
		
		// allow remaining throughput, but at least the minimum below
		else if( throughputRemaining > 0.0 ) {
			this->flows[ index ]->setThroughputLimit( std::max( throughputRemaining, throughputTarget * 0.01 ) ) ;
			throughputRemaining = 0.0 ; }
		
		// allow every processor to participate for keeping its statistics updated
		else {
			this->flows[ index ]->setThroughputLimit( throughputTarget * 0.01 ) ; }
	}
	
} // method "updateThroughputLimits()"

// print evaluation
void DemandAlloc::eval( size_t indentationName,
                        size_t indentationData ) const {
	// own statistics
	FlowBase::eval( indentationName, indentationData ) ;
	
	// find longest name of wrapped flow components
	size_t nameSizeMax = 0 ;
	for( const auto& flow : this->flows ) {
		nameSizeMax = std::max( nameSizeMax, flow->name.size() ) ;
	}
	
	// print their statistics
	for( const auto& flow : this->flows ) {
		flow->eval( indentationName + 1, nameSizeMax ) ;
	}
}

// override pulling interface to distribute final work item to all flow components
void DemandAlloc::pullImpl( ItemBase::Ptr& itemPtr ) {
	
	// keep returning final items once a single one has been seen
	if( HAWAII_UNLIKELY( this->pullFinalItem ) ) {
		itemPtr = itemFinalPtr ;
		return ;
	}
	
	// wrap base method, remember if obtained final item
	FlowBase::pullImpl( itemPtr ) ;
	if( HAWAII_UNLIKELY( itemPtr->isFinal() ) ) {
		this->pullFinalItem = true ;
	}
}
bool DemandAlloc::pullTryImpl( ItemBase::Ptr& itemPtr ) {
	
	// keep returning final items once a single one has been seen
	if( HAWAII_UNLIKELY( this->pullFinalItem ) ) {
		itemPtr = itemFinalPtr ;
		return true ;
	}
	
	// wrap base method, remember if obtained final item
	const bool success = FlowBase::pullTryImpl( itemPtr ) ;
	if( success && HAWAII_UNLIKELY( itemPtr->isFinal() ) ) {
		this->pullFinalItem = true ;
	}
	return success ;
}

// override pushing interface to retain final work item until all flow components have tried to push one
void DemandAlloc::pushImpl( const ItemBase::Ptr& itemPtr ) {
	
	// manage list of finished flow components
	if( HAWAII_UNLIKELY( itemPtr->isFinal() ) ) {
		
		// add the currently pushing thread's ID
		const boost::thread::id threadID = boost::this_thread::get_id() ;
		boost::lock_guard< boost::mutex > lock( this->pushedFinalItem.first ) ;
		this->pushedFinalItem.second.insert( threadID ) ;
		
		// do not push externally except for the last flow component
		if( this->pushedFinalItem.second.size() < this->flows.size() ) {
			return ;
		}
	}
	
	// wrap base method, optimize power every 2^n items
	FlowBase::pushImpl( itemPtr ) ;
	if( HAWAII_UNLIKELY( !this->timerUpdate.isRunning() )
	 || this->timerUpdate.toc() > this->timeUpdate        ) {
		this->timerUpdate.tic() ;
		this->updateThroughputLimits() ;
	}
}
bool DemandAlloc::pushTryImpl( const ItemBase::Ptr& itemPtr ) {
	
	// manage list of finished flow components
	if( HAWAII_UNLIKELY( itemPtr->isFinal() ) ) {
		
		// add the currently pushing thread's ID
		const boost::thread::id threadID = boost::this_thread::get_id() ;
		boost::lock_guard< boost::mutex > lock( this->pushedFinalItem.first ) ;
		this->pushedFinalItem.second.insert( threadID ) ;
		
		// do not push externally except for the last flow component
		if( this->pushedFinalItem.second.size() < this->flows.size() ) {
			return true ;
		}
	}
	
	// wrap base method, optimize power every 2^n items
	const bool ret = FlowBase::pushTryImpl( itemPtr ) ;
	if( ret ) {
		if( HAWAII_UNLIKELY( !this->timerUpdate.isRunning() )
		 || this->timerUpdate.toc() > this->timeUpdate        ) {
			this->timerUpdate.tic() ;
			this->updateThroughputLimits() ;
		}
		return true ;
	} else {
		return false ;
	}
}

} } // namespaces "hawaii::flow"
