// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// distribution of a single work item to multiple processing units
// ===============================================================

#include "hawaii/flow/partitioning.h"

namespace hawaii {
namespace flow {

// "Adapter{Src,Dst}": adapters for each wrapped flow component
// ------------------------------------------------------------

// allow the variadic template to terminate
void Partitioning::add() {
	
	// connect adapters to flow components
	{
		auto adapterIt = this->adapters.begin() ;
		auto flowPtrIt = this->flows.begin()    ;
		for( ; adapterIt != this->adapters.end() && flowPtrIt != this->flows.end() ; ++adapterIt, ++flowPtrIt ) {
			(*flowPtrIt)->setSrcIface( *adapterIt ) ;
			(*flowPtrIt)->setDstIface( *adapterIt ) ;
		}
	}
}

// implement interface for corresponding flow component
void Partitioning::Adapter::pullImpl( ItemBase::Ptr& itemPtr ) {
	
	// update performance model of corresponding flow component based on the time since the last pull finished
	// developer note: This takes less than a microsecond and therefore may be done for every work item.
	if( !trackerUpdatedBeforePullTryFailed ) {
		this->perfModelUpdate( 0.001 * this->timerThroughputInverse.toc(), this->ratioPartPullPrev ) ;
	}
	
	// take own partition from adapter, wait if unavailable
	boost::unique_lock< boost::mutex > lock( membee->mutexSrc ) ;
	while( !this->partPullPtr ) { membee->condSrcReady.wait( lock ) ; } // only difference from "pullTryImpl()"
	this->ratioPartPullPrev = this->partPullPtr->ratio ;
	itemPtr = this->partPullPtr ;
	this->partPullPtr.reset() ;
	
	// if own partition has been the last, enqueue pulling and splitting a new item (but return before that is complete)
	if( membee->adaptersSrcEmpty()
	 && HAWAII_LIKELY( !itemPtr->isFinal() ) ) {
		boost::thread pullAsync( &Partitioning::pullToAdapters, membee ) ;
		lock.unlock() ;
		pullAsync.detach() ;
	}
	
	// restart timer when the corresponding flow leaves with the work item partition
	this->trackerUpdatedBeforePullTryFailed = false ;
	this->timerThroughputInverse.tic() ;
}
bool Partitioning::Adapter::pullTryImpl( ItemBase::Ptr& itemPtr ) {
	
	// update performance model of corresponding flow component based on the time since the last pull finished
	// developer note: This takes less than a microsecond and therefore may be done for every work item.
	if( !trackerUpdatedBeforePullTryFailed ) {
		this->perfModelUpdate( 0.001 * this->timerThroughputInverse.toc(), this->ratioPartPullPrev ) ;
	}
	
	// take own partition from adapter, fail if unavailable
	boost::unique_lock< boost::mutex > lock( membee->mutexSrc ) ;
	if( !this->partPullPtr ) { this->trackerUpdatedBeforePullTryFailed = true ; return false ; } // only difference from "pullImpl()"
	this->ratioPartPullPrev = this->partPullPtr->ratio ;
	itemPtr = this->partPullPtr ;
	this->partPullPtr.reset() ;
	
	// if own partition has been the last, enqueue pulling and splitting a new item (but return before that is complete)
	if( membee->adaptersSrcEmpty()
	 && HAWAII_LIKELY( !itemPtr->isFinal() ) ) {
		boost::thread pullAsync( &Partitioning::pullToAdapters, membee ) ;
		lock.unlock() ;
		pullAsync.detach() ;
	}
	
	// restart timer when the corresponding flow leaves with the work item partition
	this->trackerUpdatedBeforePullTryFailed = false ;
	this->timerThroughputInverse.tic() ;
	
	// success
	return true ;
}
void Partitioning::Adapter::pushImpl( const ItemBase::Ptr& itemPtr ) {
	
	// give own partition to adapter, wait if destination space unavailable
	boost::unique_lock< boost::mutex > lock( membee->mutexDst ) ;
	while( this->partPushPtr ) { membee->condDstEmpty.wait( lock ) ; } // only difference from "pushTryImpl()"
	this->partPushPtr = itemPtr ;
	
	// if own partition has been the last, enqueue merging them and pushing the whole one
	if( membee->adaptersDstReady() ) {
		boost::thread pushAsync( &Partitioning::pushFromAdapters, membee ) ;
		lock.unlock() ;
		pushAsync.detach() ;
	}
}
bool Partitioning::Adapter::pushTryImpl( const ItemBase::Ptr& itemPtr ) {
	
	// give own partition to adapter, fail if destination space unavailable
	boost::unique_lock< boost::mutex > lock( membee->mutexDst ) ;
	if( this->partPushPtr ) { return false ; } // only difference from "pushImpl()"
	this->partPushPtr = itemPtr ;
	
	// if own partition has been the last, enqueue merging them and pushing the whole one
	if( membee->adaptersDstReady() ) {
		boost::thread pushAsync( &Partitioning::pushFromAdapters, membee ) ;
		lock.unlock() ;
		pushAsync.detach() ;
	}
	
	// success
	return true ;
}

// Kalman filter to estimate "x" in "1 / throughput = ratio * x[0] + x[1]" for corresponding flow component
void Partitioning::Adapter::perfModelPredict( double& throughputInverseFactor,
                                              double& throughputInverseOffset ) const {
	throughputInverseFactor = this->perfModel.x[0] ;
	throughputInverseOffset = this->perfModel.x[1] ;
}
void Partitioning::Adapter::perfModelUpdate( const double throughputInverse,
                                             const double ratio,
                                             const double noise             ) {
	// shorten names
	const double z = throughputInverse,
	             h = ratio,
	             R = noise             ;
	
	// predict state co-variance (state itself assumed constant)
	this->perfModel.P[0][0] += this->perfModel.Q[0][0] ;
	this->perfModel.P[0][1] += this->perfModel.Q[0][1] ;
	this->perfModel.P[1][0] += this->perfModel.Q[1][0] ;
	this->perfModel.P[1][1] += this->perfModel.Q[1][1] ;
	
	// compute Kalman gain
	double K[2] ;
	const double commonFactor1 = 1.0 / ( this->perfModel.P[0][0] * h * h
	                                   + this->perfModel.P[0][1] * h
	                                   + this->perfModel.P[1][0] * h
	                                   + this->perfModel.P[1][1] + R     ) ;
	K[0] = commonFactor1 * ( this->perfModel.P[0][0] * h
	                       + this->perfModel.P[0][1]     ) ;
	K[1] = commonFactor1 * ( this->perfModel.P[1][0] * h
	                       + this->perfModel.P[1][1]     ) ;
	
	// update state estimate
	const double commonFactor2 = z - ( this->perfModel.x[0] * h
	                                 + this->perfModel.x[1]     ) ;
	this->perfModel.x[0] += K[0] * commonFactor2 ;
	this->perfModel.x[1] += K[1] * commonFactor2 ;
	
	// update co-variance
	double PNew[2][2] ;
	PNew[0][0] = ( 1.0 - h * K[0] ) * this->perfModel.P[0][0] + ( 0.0 - K[0] ) * this->perfModel.P[1][0] ;
	PNew[0][1] = ( 1.0 - h * K[0] ) * this->perfModel.P[0][1] + ( 0.0 - K[0] ) * this->perfModel.P[1][1] ;
	PNew[1][0] = ( 0.0 - h * K[1] ) * this->perfModel.P[0][0] + ( 1.0 - K[1] ) * this->perfModel.P[1][0] ;
	PNew[1][1] = ( 0.0 - h * K[1] ) * this->perfModel.P[0][1] + ( 1.0 - K[1] ) * this->perfModel.P[1][1] ;
	this->perfModel.P[0][0] = PNew[0][0] ;
	this->perfModel.P[0][1] = PNew[0][1] ;
	this->perfModel.P[1][0] = PNew[1][0] ;
	this->perfModel.P[1][1] = PNew[1][1] ;
	
} // method "Partitioning::Adapter::perfModelUpdate()"

// initialize Kalman filter with default parameters
void Partitioning::Adapter::PerfModel::reset() {
	this->x[0]    = 1.0e-2 ;
	this->x[1]    = 1.0e-3 ;
	this->P[0][0] = 1.0e+1 ;
	this->P[0][1] = 0.0    ;
	this->P[1][0] = 0.0    ;
	this->P[1][1] = 1.0e+1 ;
	this->Q[0][0] = 1.0e-1 ;
	this->Q[0][1] = 0.0    ;
	this->Q[1][0] = 0.0    ;
	this->Q[1][1] = 1.0e-1 ;
}

// allow access by class whose member this is (i.e. its "membee"), know the actual instance
Partitioning::Adapter::Adapter( Partitioning* const membeeArg ) :
	timerThroughputInverse( false ),
	trackerUpdatedBeforePullTryFailed( true ),
	membee( membeeArg ) {
	this->perfModel.reset() ;
}

// "Partitioning": wrapper around identical functionality on different devices
// ---------------------------------------------------------------------------

// reset internal state
void Partitioning::reset() {
	for( auto& flow    : this->flows    ) { flow->reset() ; }
	for( auto& adapter : this->adapters ) { adapter.timerThroughputInverse.reset() ;
	                                        adapter.trackerUpdatedBeforePullTryFailed = true ;
	                                        adapter.perfModel.reset() ;                        }
	FlowBase::reset() ;
}

// wrapped flow components and their individual adapters, helper methods
bool Partitioning::adaptersSrcEmpty() const {
	bool ret = true ;
	for( const auto& adapter : this->adapters ) {
		ret &= ( adapter.partPullPtr == nullptr ) ;
	}
	return ret ;
}
bool Partitioning::adaptersDstReady() const {
	size_t numPartsReady = 0,
	       numPartsTotal = 0xFFFFFFFF ;
	for( const auto& adapter : this->adapters ) {
		if( adapter.partPushPtr ) {
			++numPartsReady ;
			HAWAII_ERROR_CONDITIONAL( numPartsTotal != 0xFFFFFFFF
			                       && numPartsTotal != adapter.partPushPtr->numParts.top(),
			                          "inconsistent total number of partitions"            ) ;
			numPartsTotal = adapter.partPushPtr->numParts.top() ;
		}
	}
	return ( numPartsReady == numPartsTotal ) ;
}

// connection between external interfaces and the above adapters
void Partitioning::pullToAdapters() {
	
	// ensure thread safety
	boost::unique_lock< boost::mutex > lock( this->mutexSrc ) ;
	
	// evaluate partitioning ratios, remove zeros
	const std::vector< double > ratiosAll = this->getRatios() ;
	      std::vector< double > ratiosUse                     ;
	for( const auto ratio : ratiosAll ) {
		if( ratio != 0.0 ) { ratiosUse.push_back( ratio ) ; }
	}
	
	// obtain and partition new item
	ItemBase::Ptr whole ;
	this->pull( whole ) ;
	ItemBase::PtrVec parts ;
	whole->split( parts, ratiosUse, *this->partInfoPtr ) ;
	
	// distribute partitions
	auto ratioIt = ratiosAll.cbegin() ;
	auto partIt  = parts.cbegin()     ;
	for( auto adapter : this->adapters ) {
		
		// skip disabled adapters
		if( *ratioIt++ == 0.0 ) { continue ; }
		
		// assign partition to adapter, set total number of partitions
		adapter.partPullPtr = *partIt++ ;
		adapter.partPullPtr->numParts.push( parts.size() ) ;
	}
	
	// notify flow components waiting to pull input
	lock.unlock() ;
	this->condSrcReady.notify_all() ;
}
void Partitioning::pushFromAdapters() {
	
	// ensure thread safety
	boost::unique_lock< boost::mutex > lock( this->mutexDst ) ;
	
	// collect partitions
	ItemBase::PtrVec parts ;
	for( auto adapterIt = this->adapters.begin() ; adapterIt != this->adapters.end() ; ++adapterIt ) {
		
		// skip disabled adapters
		if( !adapterIt->partPushPtr ) { continue ; }
		
		// get and remove partition from adapter
		parts.push_back( adapterIt->partPushPtr ) ;
		parts.back()->numParts.pop() ;
		adapterIt->partPushPtr.reset() ;
	}
	
	// merge partitions and pass on whole item
	ItemBase::Ptr whole ;
	parts.front()->merge( parts, whole, *this->partInfoPtr ) ;
	this->push( whole ) ;
	
	// notify flow components waiting to push output
	lock.unlock() ;
	this->condDstEmpty.notify_all() ;
}

// partitioning: run-time adaptation of ratios
// developer note: This takes a fraction of a millisecond and is called asynchronously, so new ratios may be computed 
//                 in every call. 
const std::vector< double > Partitioning::getRatios() {
	
	// check state
	HAWAII_ERROR_CONDITIONAL( 1.0 / this->adapters.size() < this->ratioMin,
	                          "too many partitions to maintain minimum ratio" ) ;
	
	// helper vars, pre-allocate result
	const int numFlows = this->flows.size() ;
	std::vector< double > ratios( numFlows, NAN ) ;
	bool ratiosFound = false ;
	
	// reset the performance models at regular intervals to give previously inefficient processors another chance
	if( HAWAII_UNLIKELY( !this->timerReset.isRunning() ) ) { this->timerReset.tic() ; }
	if( HAWAII_UNLIKELY( this->timerReset.toc() > this->timeReset ) ) {
		this->timerReset.tic() ;
		for( auto& adapter : this->adapters ) { adapter.timerThroughputInverse.reset() ;
		                                        adapter.trackerUpdatedBeforePullTryFailed = true ;
		                                        adapter.perfModel.reset() ;                        }
		std::fill( ratios.begin(), ratios.end(), 1.0 / (double)numFlows ) ;
		ratiosFound = true ;
	}
	
	// partition equally in case required information is not yet available
	if( HAWAII_LIKELY( !ratiosFound ) ) {
		bool noInfo = false ;
		for( auto& flow : this->flows ) {
			noInfo |= isnan( flow->ratio()      ) ;
			noInfo |= isnan( flow->throughput() ) ;
			noInfo |= isnan( flow->power()      ) ;
		}
		if( noInfo ) {
			std::fill( ratios.begin(), ratios.end(), 1.0 / (double)numFlows ) ;
			ratiosFound = true ;
		}
	}
	
	// partition to optimize power efficiency
	if( !ratiosFound && ( this->getThroughputLimit() != INFINITY
	                   || this->flowPullPtr          != nullptr  ) ) {
		
		// sort flow components by power efficiency
		std::vector< size_t > indices ; indices.reserve( this->flows.size() ) ;
		for( size_t index = 0 ; index < this->flows.size() ; ++index ) { indices.push_back( index ) ; }
		std::sort( indices.begin(), indices.end(),
		           [ this ]( size_t index1, size_t index2 ) {
			           return ( this->flows[ index1 ]->ratio() * this->flows[ index1 ]->throughput() / this->flows[ index1 ]->power()
			                  > this->flows[ index2 ]->ratio() * this->flows[ index2 ]->throughput() / this->flows[ index2 ]->power() ) ;
		           } ) ;
		
		// give most efficient flows the biggest possible ratios to still meet the throughput requirement
		double throughputInverseTarget = 1.0 / this->getThroughputLimit() ;
		if( this->flowPullPtr != nullptr
		 && this->flowPullPtr->waitedOnPush() < 0.1 ) {
			throughputInverseTarget = std::max( throughputInverseTarget,
			                                    1.0 / ( this->flowPullPtr->throughput() * 1.1 ) ) ;
		}
		double ratioRemaining = 1.0 ;
		for( const auto index : indices ) {
			
			// find ratio to exactly achieve required throughput
			double factor, offset ;
			this->adapters[ index ].perfModelPredict( factor, offset ) ;
			double ratio = ( throughputInverseTarget - offset ) / factor ;
			
			// requested throughput cannot be achieved
			if( ratio < 0.0 && ratioRemaining > 0.0 ) { break ; }
			
			// set ratio and continue
			ratios[ index ] = ratio = std::max( 0.0, std::min( ratioRemaining, ratio ) ) ;
			ratioRemaining = std::max( ratioRemaining - ratio, 0.0 ) ;
		}
		
		// if power-efficient partitioning failed, switch to maximum performance mode
		if( ratioRemaining == 0.0 ) {
			ratiosFound = true ;
			
			// print model parameters and results
			#if 0
				hawaii::cout << "partitioning for efficiency" << hawaii::endl ;
				for( size_t index = 0 ; index < this->adapters.size() ; ++index ) {
					double factor, offset ;
					this->adapters[ index ].perfModelPredict( factor, offset ) ;
					hawaii::cout << "factor = " << factor << ", offset = " << offset << ", ratio = " << ratios[ index ] << hawaii::endl ;
				}
			#endif
		}
	}
	
	// partition to achieve same throughput for all flow components, i.e. maximize performance
	if( !ratiosFound ) {
		
		// For each flow, the performance model means "throughput = ratio * factor + offset". For "n" flow components, we 
		// set up a system of linear equations "A * x = b" according to:
		//  ( factor #1  -factor #2      ...         0           0     )   ( ratio #1   )   ( offset #2 - offset #1   )
		//  (     0       factor #2      ...         0           0     )   ( ratio #2   )   (           ...           )
		//  (    ...         ...         ...        ...         ...    ) * (    ...     ) = (           ...           )
		//  (     0           0          ...    factor #n-1  factor #n )   ( ratio #n-1 )   ( offset #n - offset #n-1 )
		//  (     1           1          ...         1           1     )   ( ratio #n   )   (            1            )
		// Equation "k, n < n" state that flow #k and #k+1 should achieve the same throughput; "k" states that the sum of 
		// all partition ratios should be 1.
		cv::Mat_< double > A( numFlows, numFlows, 0.0 ),
		                   b( numFlows, 1       , 0.0 ),
		                   x( numFlows, 1       , 0.0 ) ;
		for( int n = 0 ; n < numFlows - 1 ; ++n ) {
			double factor, offset ;
			this->adapters[ n     ].perfModelPredict( factor, offset ) ;
			A( n, n     ) += factor ;
			b( n        ) -= offset ;
			this->adapters[ n + 1 ].perfModelPredict( factor, offset ) ;
			A( n, n + 1 ) -= factor ;
			b( n        ) += offset ;
		}
		A.row( numFlows - 1 ) = b( numFlows - 1 ) = 1.0 ;
		cv::solve( A, b, x ) ;
		
		// limit ratios to [0,1]: This might not always be the case, e.g. if the "offset" throughput of a slow flow is 
		//                        higher than the "offset + factor * 1.0" throughput of a fast one.
		for( int n = 0 ; n < numFlows ; ++n ) {
			ratios[ n ] = std::max( 0.0, std::min( 1.0, x( n ) ) ) ;
		}
		
		// print model parameters and results
		#if 0
			hawaii::cout << "partitioning for performance" << hawaii::endl ;
			for( size_t index = 0 ; index < this->adapters.size() ; ++index ) {
				double factor, offset ;
				this->adapters[ index ].perfModelPredict( factor, offset ) ;
				hawaii::cout << "factor = " << factor << ", offset = " << offset << ", ratio = " << ratios[ index ] << hawaii::endl ;
			}
		#endif
	}
	
	// disable processors with ratios below minimum
	double sumBelow = 0.0,
	       sumAbove = 0.0 ;
	for( const auto& ratio : ratios ) {
		if( ratio < this->ratioMin ) { sumBelow += this->ratioMin - ratio ; }
		else                         { sumAbove += ratio - this->ratioMin ; }
	}
	for( auto& ratio : ratios ) {
		if( ratio < this->ratioMin ) { ratio  = 0.0                                              ; }
		else                         { ratio += ( ratio - this->ratioMin ) / sumAbove * sumBelow ; }
	}
	
	// done
	return ratios ;
}

// run this partitioner, wait for it to finish or directly stop it
void Partitioning::run()  { this->stop() ;
                            this->reset() ;
                            if( this->adaptersSrcEmpty() ) { this->pullToAdapters() ; }
                            for( FlowBase::Ptr& flow : this->flows ) { flow->run()  ; } }
void Partitioning::wait() { for( FlowBase::Ptr& flow : this->flows ) { flow->wait() ; } }
void Partitioning::stop() { for( FlowBase::Ptr& flow : this->flows ) { flow->stop() ; } }

// estimate power consumption
double Partitioning::power() const {
	double ret = 0.0 ;
	for( const auto& flow : this->flows ) {
		ret += flow->power() ;
	}
	return ret ;
}

// print evaluation
void Partitioning::eval( size_t indentationName,
                         size_t indentationData ) const {
	// own statistics
	FlowBase::eval( indentationName, indentationData ) ;
	
	// find longest name of wrapped flow components
	size_t nameSizeMax = 0 ;
	for( const auto& flow : this->flows ) {
		nameSizeMax = std::max( nameSizeMax, flow->name.size() ) ;
	}
	
	// print their statistics
	for( const auto& flow : this->flows ) {
		flow->eval( indentationName + 1, nameSizeMax ) ;
	}
}

// safely destruct mutexes and conditions
Partitioning::~Partitioning() {
	this->wait() ;
	boost::lock_guard< boost::mutex > lockSrc( this->mutexSrc ) ;
	boost::lock_guard< boost::mutex > lockDst( this->mutexDst ) ;
}

} } // namespaces "hawaii::flow"
