// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// queues to buffer work items
// ===========================

#include "hawaii/flow/queue.h"

namespace hawaii {
namespace flow {

// factory function for any of the queues below
QueueBase* queueFactory( const QueueType queueTypeArg,
                         const size_t    capacityArg  ) {
	
	switch( queueTypeArg ) {
	case queueTypeFIFO             : return new QueueFIFO(             capacityArg ) ; break ;
	case queueTypeFIFOOverflow     : return new QueueFIFOOverflow(     capacityArg ) ; break ;
	case queueTypePriority         : return new QueuePriority        ( capacityArg ) ; break ;
	case queueTypePriorityOverflow : return new QueuePriorityOverflow( capacityArg ) ; break ;
	case queueTypeOrdered          : return new QueueOrdered(          capacityArg ) ; break ;
	case queueTypeOrderedOverflow  : return new QueueOrderedOverflow(  capacityArg ) ; break ;
	default:
		HAWAII_ERROR( "unknown queue type requested" ) ;
		return nullptr ;
		break ;
	}
}

// "QueueBase": common part of all queues
// --------------------------------------

// inherited push/pull interface, override only if necessary
void QueueBase::pushImpl( const ItemBase::Ptr& itemPtr ) {
	boost::unique_lock< Mutex > lock( this->mutex ) ;
	while( !this->pushTryUnlocked( itemPtr ) ) {
		this->pushable.wait( lock ) ;
	}
	lock.unlock() ;
	this->pullable.notify_one() ;
}
bool QueueBase::pushTryImpl( const ItemBase::Ptr& itemPtr ) {
	this->mutex.lock() ;
	const bool success = this->pushTryUnlocked( itemPtr ) ;
	this->mutex.unlock() ;
	if( success ) { this->pullable.notify_one() ; }
	return success ;
}
void QueueBase::pullImpl( ItemBase::Ptr& itemPtr ) {
	boost::unique_lock< Mutex > lock( this->mutex ) ;
	while( !this->pullTryUnlocked( itemPtr ) ) {
		this->pullable.wait( lock ) ;
	}
	lock.unlock() ;
	this->pushable.notify_one() ;
}
bool QueueBase::pullTryImpl( ItemBase::Ptr& itemPtr ) {
	this->mutex.lock() ;
	const bool success = this->pullTryUnlocked( itemPtr ) ;
	this->mutex.unlock() ;
	if( success ) { this->pushable.notify_one() ; }
	return success ;
}
	
// "QueueFIFOOverflow": FIFO queue with overflow
// ---------------------------------------------
	
// simply wrap "boost::circular_buffer", which behaves exactly as described above
bool QueueFIFOOverflow::pushTryUnlocked( const ItemBase::Ptr& itemPtr ) {
	this->impl.push_back( itemPtr ) ;
	return true ;
}
bool QueueFIFOOverflow::pullTryUnlocked( ItemBase::Ptr& itemPtr ) {
	if( this->emptyUnlocked() ) {
		return false ;
	} else {
		itemPtr = this->impl.front() ;
		this->impl.pop_front() ;
		return true ;
	}
}

// "QueueFIFO": FIFO queue without overflow
// ----------------------------------------

// change push interface to from overflow to rejecting
bool QueueFIFO::pushTryUnlocked( const ItemBase::Ptr& itemPtr ) {
	if( this->fullUnlocked() ) {
		return false ;
	} else {
		this->impl.push_back( itemPtr ) ;
		return true ;
	}
}

// "QueuePriorityOverflow": allow items with a lower serial number to "overtake" others, with overflow like "QueueFIFOOverflow"
// ----------------------------------------------------------------------------------------------------------------------------

// "std::priority_queue" does the sorting, but not the overflow.
bool QueuePriorityOverflow::pushTryUnlocked( const ItemBase::Ptr& itemPtr ) {
	this->impl.push( itemPtr ) ;
	while( this->sizeUnlocked() > this->capacityUnlocked() ) {
		this->impl.pop() ;
	}
	return true ;
}
bool QueuePriorityOverflow::pullTryUnlocked( ItemBase::Ptr& itemPtr ) {
	if( this->emptyUnlocked() ) {
		return false ;
	} else {
		itemPtr = this->impl.top() ;
		this->impl.pop() ;
		return true ;
	}
}

// "QueuePriority": allow items with a lower serial number to "overtake" others, without overflow like "QueueFIFO"
// ---------------------------------------------------------------------------------------------------------------

// change push interface to from overflow to rejecting
bool QueuePriority::pushTryUnlocked( const ItemBase::Ptr& itemPtr ) {
	if( this->fullUnlocked() ) {
		return false ;
	} else {
		this->impl.push( itemPtr ) ;
		return true ;
	}
}

// "QueueOrderedOverflow": offer items to be pulled ordered by serial number
// -------------------------------------------------------------------------

// change push interface to drop "out-dated" items on arrival
bool QueueOrderedOverflow::pushTryUnlocked( const ItemBase::Ptr& itemPtr ) {
	// developer note: ">=" instead of ">" so that item #0 won't be dropped and, if serial numbers are identical 
	//                 for whatever reason, the queue won't drop all of them but instead only degenerate to a 
	//                 "QueueFIFOOverflow".
	if( itemPtr->serialNumber >= this->serialNumberPulled ) {
		this->impl.push( itemPtr ) ;
		while( this->sizeUnlocked() > this->capacityUnlocked() ) {
			this->impl.pop() ;
		}
	}
	return true ;
}
	
// change pull interface to update last pulled serial number
bool QueueOrderedOverflow::pullTryUnlocked( ItemBase::Ptr& itemPtr ) {
	if( this->emptyUnlocked() ) {
		return false ;
	} else {
		itemPtr = this->impl.top() ;
		this->impl.pop() ;
		this->serialNumberPulled = itemPtr->serialNumber ;
		return true ;
	}
}

// "QueueOrdered": offer items to be pulled with consecutive serial numbers
// ------------------------------------------------------------------------

// change push interface to reject items with too high serial numbers: If e.g. #5 must be pulled next and capacity 
// is 4, only accept #5-8 to be pushed; #9 must wait until #5 is out again.
bool QueueOrdered::pushTryUnlocked( const ItemBase::Ptr& itemPtr ) {
	if( (int64_t)( itemPtr->serialNumber - this->serialNumberNext ) >= (int64_t)this->capacityUnlocked()
	 && HAWAII_LIKELY( !itemPtr->isFinal() ) ) {
		return false ;
	} else {
		this->impl.push( itemPtr ) ;
		return true ;
	}
}

// change pull interface to block or fail if the next consecutive item is not here yet
bool QueueOrdered::pullTryUnlocked( ItemBase::Ptr& itemPtr ) {
	if( this->emptyUnlocked()
	 || ( this->impl.top()->serialNumber != this->serialNumberNext
	   && HAWAII_LIKELY( !this->impl.top()->isFinal() ) ) ) {
		return false ;
	} else {
		itemPtr = this->impl.top() ;
		this->impl.pop() ;
		if( HAWAII_LIKELY( !itemPtr->isFinal() ) ) {
			++this->serialNumberNext ;
		}
		return true ;
	}
}
	
// also override public pull interface to notify all instead of just one pushing thread, because the one's item may 
// be rejected 
void QueueOrdered::pullImpl( ItemBase::Ptr& itemPtr ) {
	boost::unique_lock< Mutex > lock( this->mutex ) ;
	while( !this->pullTryUnlocked( itemPtr ) ) {
		this->pullable.wait( lock ) ;
	}
	lock.unlock() ;
	this->pushable.notify_all() ;
}
bool QueueOrdered::pullTryImpl( ItemBase::Ptr& itemPtr ) {
	this->mutex.lock() ;
	const bool success = this->pullTryUnlocked( itemPtr ) ;
	this->mutex.unlock() ;
	if( success ) { this->pushable.notify_all() ; }
	return success ;
}

} } // namespaces "hawaii::flow"
