// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// base classes for GPU processing steps to overlap computations and memory copies
// ===============================================================================

#include "hawaii/flow/stepGPU.h"
#include "hawaii/GPU/CUDA.h"
#include "hawaii/common/hardware.h"
#include <boost/thread/thread.hpp>
#include <sys/time.h>
#include <sys/resource.h>

// helpers only used here
namespace {
	
	// estimate power consumption every "itemsPerUpdate" items (must be at least 3!)
	constexpr uint64_t itemsPerPowerUpdate  = 4,
	                   itemsPerPowerUpdate1 = itemsPerPowerUpdate - 1 ; 
	
	// get the CPU time (user and system) in [cores * micro-seconds] spent by this thread
	// user note: This does not see OpenMP threads spawned between calls!
	// developer note #1: "boost::cpu_timer" is not available before version 1.48. As long as 1.46 is still widely used, 
	//                    e.g. via Ubuntu 12.04, we skip portability and do this manually.
	// developer note #2: A copy of this code exists in "src/flow/step.cpp" - keep them in sync!
	uint64_t getSpentTimeCPU() {
		rusage temp ;
		getrusage( RUSAGE_THREAD, &temp ) ;
		return temp.ru_utime.tv_sec * 1000000 + temp.ru_utime.tv_usec
		     + temp.ru_stime.tv_sec * 1000000 + temp.ru_stime.tv_usec ;
	}
	
	// wrap "StepBaseGPU::{upload,compute,download}()" to include handling the final item
	inline void uploadHelper(       hawaii::flow::StepBaseGPU* const that,
	                          const hawaii::flow::ItemBase::Ptr&     srcPtr,
	                                hawaii::flow::ItemBase::Ptr&     srcGPUPtr,
	                                cv::gpu::Stream&                 stream    ) {
		
		// do not really upload final item, as it may really be only a "hawaii::flow::ItemBase"
		if( HAWAII_LIKELY( !srcPtr->isFinal() ) ) {
			that->upload( srcPtr, srcGPUPtr, stream ) ;
		} else {
			srcGPUPtr = hawaii::flow::itemFinalPtr ;
		}
	}
	inline void computeHelper(       hawaii::flow::StepBaseGPU* const that,
	                           const hawaii::flow::ItemBase::Ptr&     srcGPUPtr,
	                                 hawaii::flow::ItemBase::Ptr&     dstGPUPtr,
	                                 cv::gpu::Stream&                 stream,
	                           const uint64_t                         countdownPowerUpdate,
	                                 cudaEvent_t&                     eventGPUStart,
	                                 cudaEvent_t&                     eventGPUEnd          ) {
		
		// do not really process final item, as it may really be only a "hawaii::flow::ItemBase"
		if( HAWAII_LIKELY( !srcGPUPtr->isFinal() ) ) {
			
			// enqueue taking time when GPU(s) can(!) start working, assuming that they also will
			#ifndef HAWAII_NO_CUDA
				if( countdownPowerUpdate == 3 ) { HAWAII_ERROR_CHECK_CUDA( cudaEventRecord( eventGPUStart, cv::gpu::StreamAccessor::getStream( stream ) ) ) ; }
			#else
				(void)countdownPowerUpdate ;
				(void)eventGPUStart        ;
			#endif
			
			// enqueue the processing
			that->compute( srcGPUPtr, dstGPUPtr, stream ) ;
			
			// enqueue taking the time GPU(s) is/are done
			#ifndef HAWAII_NO_CUDA
				if( countdownPowerUpdate == 3 ) { HAWAII_ERROR_CHECK_CUDA( cudaEventRecord( eventGPUEnd,   cv::gpu::StreamAccessor::getStream( stream ) ) ) ; }
			#else
				(void)countdownPowerUpdate ;
				(void)eventGPUEnd          ;
			#endif
			
		} else {
			dstGPUPtr = hawaii::flow::itemFinalPtr ;
		}
	}
	inline void downloadHelper(       hawaii::flow::StepBaseGPU* const that,
	                            const hawaii::flow::ItemBase::Ptr&     dstGPUPtr,
	                                  hawaii::flow::ItemBase::Ptr&     dstPtr,
	                                  cv::gpu::Stream&                 stream    ) {
		
		// do not really download final item, as it may really be only a "hawaii::flow::ItemBase"
		if( HAWAII_LIKELY( !dstGPUPtr->isFinal() ) ) {
			that->download( dstGPUPtr, dstPtr, stream ) ;
		} else {
			dstPtr = hawaii::flow::itemFinalPtr ;
		}
	}
	
} // anonymous namespace

namespace hawaii {
namespace flow {

// "StepBaseGPU": (really) base GPU processing step
// ------------------------------------------------

// processing function to wrap "StepBaseGPU::{upload,compute,download}()"
void StepBaseGPU::process( const ItemBase::Ptr& srcBasePtr,
                                 ItemBase::Ptr& dstBasePtr ) {
	// intermediate work items
	ItemBase::Ptr srcBaseGPUPtr, dstBaseGPUPtr ;
	
	// enqueue all parts of processing in an individual non-blocking stream and wait for its completion
	this->activateGPU() ;
	cv::gpu::Stream stream = hawaii::GPU::streamNonBlockingFactory() ;
	this->upload(   srcBasePtr,    srcBaseGPUPtr, stream ) ;
	this->compute(  srcBaseGPUPtr, dstBaseGPUPtr, stream ) ;
	this->download( dstBaseGPUPtr, dstBasePtr,    stream ) ;
	stream.waitForCompletion() ;
}

// c'tor with one specific GPU to run on
// user note: A GPU step can be neither a source nor a sink.
StepBaseGPU::StepBaseGPU( const std::string nameArg,
                          const int         GPUArg  ) :
	StepBase( nameArg,
				 1, GPUArg,
				 false, false ),
	overlapKernelsAndCopies( true ),
	counterItemsNoOverlap( 0 ) {
}

// print evaluation
// developer note: apply changes to "FlowBase::eval()" as well
void StepBaseGPU::eval( size_t indentationName,
                        size_t indentationData ) const {
	// prepare
	indentationData = std::max( indentationData, this->name.size() ) ;
	const std::string idtName(      indentationName * 4,                 ' ' ) ;
	const std::string idtDataFirst( indentationData - this->name.size(), ' ' ) ;
	const std::string idtDataOther( indentationData + 2,                 ' ' ) ;
	
	// print
	std::ostringstream noteOverlap, lineRatio ;
	if( this->overlapKernelsAndCopies ) {
		noteOverlap << " (skipped overlap for " << this->counterItemsNoOverlap << ")" ;
	}
	if( this->ratio() != 1.0 ) {
		lineRatio << idtName                       << idtDataOther << "ratio                      = " << this->ratio() * 100.0              << " %\n" ;
	}
	hawaii::cout << idtName << this->name << ": " << idtDataFirst << "latency                    = " << this->latency()                    << " ms\n"
	             << idtName                       << idtDataOther << "throughput                 = " << this->throughput()                 << " Hz\n"
	             << idtName                       << idtDataOther << "throughput (unconstrained) = " << this->throughputUnconstrained()    << " Hz\n"
	             << idtName                       << idtDataOther << "power                      = " << this->power()                      << " W\n"
	             << idtName                       << idtDataOther << "items                      = " << this->items() << noteOverlap.str() << "\n"
	             << idtName                       << idtDataOther << "waited on pull             = " << this->waitedOnPull() * 100.0       << " %\n"
	             << idtName                       << idtDataOther << "waited on push             = " << this->waitedOnPush() * 100.0       << " %\n"
	             << lineRatio.str()
	             << hawaii::endl ;
}

// run this processing step
// developer note: Even though it says "StepBaseGPU::threadFunc", "run()" will use a derived class's overriding 
//                 implementation if available.
void StepBaseGPU::run() {
	this->stop() ;
	this->reset() ;
	if( this->overlapKernelsAndCopies
	 && hawaii::GPUMemoryOwn( this->getGPU() ) ) {
		this->thread = boost::thread( &StepBaseGPU::threadFuncGPU, this ) ;
	} else {
		this->thread = boost::thread( &StepBaseGPU::threadFunc,    this ) ;
	}
}

// method to run within this processing step
void StepBaseGPU::threadFuncGPU() {
	
	// activate selected GPU
	this->activateGPU() ;
	
	// helper vars: Index 0/1/2 refers to item pulled(!) at this/last/second-to-last iteration.
	// dev. note #1: wrap streams on stack to have them deleted on interruption
	// dev. note #2: pretend that non-existent items in first two iterations already have been pushed and timed
	cv::gpu::Stream  streamInstances[ 3 ] = { hawaii::GPU::streamNonBlockingFactory(),
	                                          hawaii::GPU::streamNonBlockingFactory(),
	                                          hawaii::GPU::streamNonBlockingFactory() } ;
	cv::gpu::Stream* streams[ 3 ] = { &streamInstances[ 0 ],
	                                  &streamInstances[ 1 ],
	                                  &streamInstances[ 2 ] } ;
	bool pushed[ 3 ] = { false, true, true } ;
	ItemBase::Ptr srcPtrs[    3 ],
	              srcGPUPtrs[ 3 ],
	              dstGPUPtrs[ 3 ],
	              dstPtrs[    3 ] ;
	
	// user and system CPU time spent by this step's thread, counter for missed updates (The information from the 
	// system call behind "getSpentTimeCPU()" sometimes does not update fast enough.)
	// developer note: A copy of this code exists in "src/flow/step.cpp" - keep them in sync!
	uint64_t spentTimeCPUsPrev = getSpentTimeCPU() ;
	uint64_t powerUpdatesSkipped = 0 ;
	
	// GPU time spent while "compute()" is called
	// developer note: A copy of this code exists in "src/flow/step.cpp" - keep them in sync!
	cudaEvent_t eventGPUStart, eventGPUEnd ;
	#ifndef HAWAII_NO_CUDA
		HAWAII_ERROR_CHECK_CUDA( cudaEventCreate( &eventGPUStart, cudaEventDefault ) ) ;
		HAWAII_ERROR_CHECK_CUDA( cudaEventCreate( &eventGPUEnd,   cudaEventDefault ) ) ;
		double GPUsPower = 0.0 ;
		for( const auto GPU : this->getGPUs() ) {
			GPUsPower += hawaii::GPUPower( GPU ) ;
		}
	#endif
	
	// decide about doing power consumption update (later)
	uint64_t countdownPowerUpdate = 0 ;
	
	// for catching a "boost::thread_interrupted"
	try {
	
		// stop via interruption point instead of loop logic
		while( true ) {
			
			// decide about doing power consumption update: enqueue now, evaluate two loops later
			if( this->counterItems % itemsPerPowerUpdate == itemsPerPowerUpdate1 ) { countdownPowerUpdate = 3 ; }
			
			// race between conditions A) and B) below
			while( true ) {
				
				// along the way, try to push item #n-2 ASAP but don't yet wait for it
				if( !pushed[ 2 ] && streams[ 2 ]->queryIfComplete() ) {
					pushed[ 2 ] = this->pushTry( dstPtrs[ 2 ] ) ;
				}
				
				// A) if item #n becomes available first, overlap copies and kernels
				if( this->pullTry( srcPtrs[ 0 ] ) ) {
					
					// GeForce: Upload #n can overlap with kernel #n-1 from the previous iteration, but must wait for 
					//          download #n-2 from two iteration ago to finish.
					// Tesla: Upload #n can overlap with both kernel #n-1 from one and download #n-2 from two iterations ago.
					uploadHelper( this, srcPtrs[ 0 ], srcGPUPtrs[ 0 ], *streams[ 0 ] ) ;
					
					// Kernel #n and download #n-1 can always overlap and start once upload #n and kernel #n-1 have finished 
					// respectively. Their order is unimportant except for cards with both two copy engines (i.e. Tesla) and 
					// concurrent kernels support (i.e. since Fermi): Download #n-1 must then be enqueued first in order not 
					// to wait unnecessarily for kernel #n to complete.
					// GeForce: Upload #n+1 from the next iteration must wait for download #n-1 to finish, but can then 
					//          overlap with kernel #n also.
					// Tesla: Upload #n+1 from the next iteration can overlap with both kernel #n and download #n-1.
					if( !pushed[ 1 ] ) {
						downloadHelper( this, dstGPUPtrs[ 1 ], dstPtrs[ 1 ], *streams[ 1 ] ) ;
					}
					computeHelper( this, srcGPUPtrs[ 0 ], dstGPUPtrs[ 0 ], *streams[ 0 ], countdownPowerUpdate, eventGPUStart, eventGPUEnd ) ;
					
					// race won by A)
					break ;
					
				} // if "condition A)?"
				
				// B) if computation #n-1 finishes first, abandon overlap
				if( streams[ 1 ]->queryIfComplete() ) {
					
					// enqueue download #n-1 earlier than usual even though it won't overlap with anything previously 
					// enqueued, count if this occurs
					if( !pushed[ 1 ] ) {
						downloadHelper( this, dstGPUPtrs[ 1 ], dstPtrs[ 1 ], *streams[ 1 ] ) ;
						++this->counterItemsNoOverlap ;
					}
					
					// race between conditions C) and D) below
					while( true ) {
						
						// along the way, try to push item #n-2 ASAP but don't yet wait for it
						if( !pushed[ 2 ] && streams[ 2 ]->queryIfComplete() ) {
							pushed[ 2 ] = this->pushTry( dstPtrs[ 2 ] ) ;
						}
						
						// C) if item #n becomes available first, enqueue its upload and kernel
						if( this->pullTry( srcPtrs[ 0 ] ) ) {
							
							// GeForce: Upload #n must wait for download #n-1 above to finish.
							// Tesla: Upload #n and download #n-1 above overlap with each other, but without any kernel.
							uploadHelper(  this, srcPtrs[    0 ], srcGPUPtrs[ 0 ], *streams[ 0 ] ) ; 
							computeHelper( this, srcGPUPtrs[ 0 ], dstGPUPtrs[ 0 ], *streams[ 0 ], countdownPowerUpdate, eventGPUStart, eventGPUEnd ) ;
							
							// race won by C)
							break ;
						}
						
						// D) if both item #n-2 has been pushed and download #n-1 finishes first, try to also push item #n-1
						if( pushed[ 2 ] && !pushed[ 1 ] && streams[ 1 ]->queryIfComplete() ) {
							pushed[ 1 ] = this->pushTry( dstPtrs[ 1 ] ) ;
							
							// race won by D), see logic below for the expected "break"
						}
						
						// if all previous items have been pushed...
						if( pushed[ 2 ] && pushed[ 1 ] ) {
							
							// avoid infinite waiting after final item 
							if( HAWAII_LIKELY(  dstPtrs[ 1 ] == nullptr
							                || !dstPtrs[ 1 ]->isFinal() ) ) {
								
								// finally block until item #n has become available
								this->pull( srcPtrs[ 0 ] ) ;
								
								// enqueue its upload and kernel, won't overlap with anything previously enqueued
								uploadHelper(  this, srcPtrs[    0 ], srcGPUPtrs[ 0 ], *streams[ 0 ] ) ; 
								computeHelper( this, srcGPUPtrs[ 0 ], dstGPUPtrs[ 0 ], *streams[ 0 ], countdownPowerUpdate, eventGPUStart, eventGPUEnd ) ;
							}
							
							// race won by D)
							break ;
						}
						
						// avoid spinning
						boost::this_thread::yield() ;
						
					} // while "race C) vs D) still going on?"
					
					// race won by B)
					break ;
					
				} // if "condition B)?"
				
				// avoid spinning
				boost::this_thread::yield() ;
				
			} // while "race A) vs B) still going on?"
			
			// if item #n-2 has not already been pushed, finally wait for it
			if( !pushed[ 2 ] ) {
				streams[ 2 ]->waitForCompletion() ;
				this->push( dstPtrs[ 2 ] ) ;
				pushed[ 2 ] = true ;
			}
			
			// estimate power consumption, assuming linear relations with each processor's thermal design power
			if( --countdownPowerUpdate == 0 ) {
				
				// compute user and system CPU time spent since last time
				const uint64_t spentTimeCPUsCurr = getSpentTimeCPU() ;
				const uint64_t spentTimeCPUsDelta = spentTimeCPUsCurr - spentTimeCPUsPrev ;
				if( spentTimeCPUsDelta > 0 ) {
					
					// this processing step's cycle time in milli-seconds
					const double msCycle           = 1000.0 / this->throughput() ;
					const double msSinceLastUpdate = msCycle * itemsPerPowerUpdate * ( 1 + powerUpdatesSkipped ) ;
					
					// compute CPU usage ratio: extrapolate to user-specified number of used cores, normalize depending on 
					//                          hyper-threading
					const double ratioCPUs = spentTimeCPUsDelta * 0.001 / msSinceLastUpdate
					                       * (double)this->CPUCores / ( this->CPUCores > hawaii::CPUCores ?
					                     			                      (double)hawaii::CPUThreads        :
					                     			                      (double)hawaii::CPUCores          ) ;
					double power = ratioCPUs * hawaii::CPUPower ;
					
					// compute GPU usage ratio
					#ifndef HAWAII_NO_CUDA
						if( GPUsPower != 0.0 ) {
							HAWAII_ERROR_CHECK_CUDA( cudaEventSynchronize( eventGPUEnd ) ) ;
							float msGPUs ; HAWAII_ERROR_CHECK_CUDA( cudaEventElapsedTime( &msGPUs, eventGPUStart, eventGPUEnd ) ) ;
							const double ratioGPUs = (double)msGPUs / msCycle ;
							power += ratioGPUs * GPUsPower ;
						}
					#endif
					
					// update tracker, prepare next measurement
					this->trackerPower( power ) ;
					spentTimeCPUsPrev = spentTimeCPUsCurr ;
					powerUpdatesSkipped = 0 ;
				}
				
				// workaround in case timers have not yet been updates by the OS
				else { ++powerUpdatesSkipped ; }
			}
			
			// end thread on final item, push it if necessary
			if( HAWAII_UNLIKELY( dstPtrs[ 1 ] != nullptr
			                  && dstPtrs[ 1 ]->isFinal() ) ) {
				if( !pushed[ 1 ] ) {
					streams[ 1 ]->waitForCompletion() ;
					this->push( dstPtrs[ 1 ] ) ;
					pushed[ 1 ] = true ;
				}
				break ;
			}
			
			// allow thread to be interrupted at least once per loop
			boost::this_thread::interruption_point() ;
			
			// update helper vars for next iteration
			cv::gpu::Stream* stream2 = streams[ 2 ] ;
			streams[    2 ] = streams[    1 ] ; streams[    1 ] = streams[    0 ] ; streams[    0 ] = stream2 ;
			pushed[     2 ] = pushed[     1 ] ; pushed[     1 ] = pushed[     0 ] ; pushed[     0 ] = false ;
			srcPtrs[    2 ] = srcPtrs[    1 ] ; srcPtrs[    1 ] = srcPtrs[    0 ] ; srcPtrs[    0 ] = ItemBase::Ptr() ;
			srcGPUPtrs[ 2 ] = srcGPUPtrs[ 1 ] ; srcGPUPtrs[ 1 ] = srcGPUPtrs[ 0 ] ; srcGPUPtrs[ 0 ] = ItemBase::Ptr() ;
			dstGPUPtrs[ 2 ] = dstGPUPtrs[ 1 ] ; dstGPUPtrs[ 1 ] = dstGPUPtrs[ 0 ] ; dstGPUPtrs[ 0 ] = ItemBase::Ptr() ;
			dstPtrs[    2 ] = dstPtrs[    1 ] ; dstPtrs[    1 ] = dstPtrs[    0 ] ; dstPtrs[    0 ] = ItemBase::Ptr() ;
			
		} // while "thread not interrupted?"
	
	// end in a joinable way
	} catch( boost::thread_interrupted& ) { /* just continue */ }
	
	// GPU time spent while "compute()" is called
	// developer note: A copy of this code exists in "src/flow/step.cpp" - keep them in sync!
	#ifndef HAWAII_NO_CUDA
		HAWAII_ERROR_CHECK_CUDA( cudaEventDestroy( eventGPUStart ) ) ;
		HAWAII_ERROR_CHECK_CUDA( cudaEventDestroy( eventGPUEnd   ) ) ;
	#endif
	
} // method "StepBaseGPU::threadFunc()"

} } // namespaces "hawaii::flow"
